//===============================================================
//
//       SakIMAP unit, part of SakEmail
//
//       Contains: TSakIMAP
//
//---------------------------------------------------------------
//
//      Copyrigth (c) 1997, 1998, 1999, 2000, 2001 Sergio A. Kessler
//      and authors cited
//      http://sak.org.ar
//
//===============================================================


unit SakIMAP;

interface

uses SysUtils, Classes, scktcomp, sak_util, SakMsg;

const
   SAK_IMAP_ERROR_CONST = 666;   // === 2001/12/20 - As formerly defined in FDoErr ===

type

  // === Comments by: James.
  // === 2001/03/22 - Utilize the TIMAPFlag enumerated type to compress "code bloat" of
  // ===        the previous quick and dirty implementation of multi-method flag manipulation.
  // ===        The new implementation differs from StoreFlags in that you may submit 
  // ===        set or clear requests for a specific flag without having to worry about
  // ===        manipulating sets.  A later step may be to add duplicate methods that allow
  // ===        utilization of the UID for specifying message behaviour instead of MsgIndex.
  // ===   With set and clear methods for each there are a total of 4 methods available.
  // ===   For convenience the structure is TOnSetFlagEvent and TOnClearFlagEvent for
  // ===        setting and clearing flag structures.  These events might be useful
  // ===        for a listener type relationship at some point.  The exception to the typical behaviour
  // ===        is for the delete method because it already exists.  In this case when TOnDeleteEvent
  // ===        handlers are called, the TOnSetFlagEventformat is also called.
  // ===        To minimize the number of events ( and fall in line with this component's code conventions ) there
  // ===        is a set and clear implementation using only 2 event types with the TIMAPFlag enumerated type.
  // ===   For further convenience an UndeleteMessage method is provided.
  // ===   A slightly cleaner interface to the ACL lists is also implemented using the
  // ===        TIMAPACLRight enumerated type below.
  // ===   The ACL extensions are as described in RFC2086.
  // === 2001/12/10 - Provided Operation timeout for non-responding receives.  Added abend to force disconnect/reset.
  // ===        Provided folder name fix based on an error discovered by Emmanuel Mazy.
  // === 2001/12/11 - RFC2060 compliancy "issue" found.  The folderlist includes inacessible folders.
  // ===        For convenience new AvFolderlist ( available ) and AvSUBFolderList ( available ) have been added and
  // ===        are populated without including unavailable folders.  Fix based on an error discovered by
  // ===        Emmanuel Mazy.
  // === 2001/12/28 - Added virtually all other functions necessary to round out RFC2060 as well as
  // ===        RFC2177 IDLE extension.  Improvements in performance and timeout safety have been added as
  // ===        well - James.
  // === 2001/12/31 - Testing and bugfixes for the new methods has been performed.  Performance behaviour
  // ===        is a compromise between performance and responsiveness during IDLE.  Some overlapping calls
  // ===        replaced with the new "lowlevel" methods ( fetch, copy, search and store ).
  // === 2002/01/02 - Adding RFC2342 Namespace query command.  Nothing fancy, but allows the
  // ===        client to get namespace information by checking ReplyString ( if supported by server ).
  // ===        Also added hierarchy control ( % or * by default ) and explicit LIST command.  Set
  // ===        ListFullHierarchy = False to use % for folder lists, etc instead of default *.
  // === 2002/01/03 - Removed compiler warnings/hints concerning unused vars/returns.
  // ===        Implemented and extended Holger Mauermann's solution for folder SELECT
  // ===        in the SelectFolder method.  It should now work with imapd 2001a.

  // 2001/03/21 - IMAP ACL Rights - NOTE: Post is not utilized by IMAP but referenced for completeness - James.
  TIMAPACLRight = (imapLookup, imapRead, imapSeen, imapWrite, imapInsert, imapPost, imapCreate, imapDelete, imapAdminister);
  TIMAPACLRights = set of TIMAPACLRight;
  // 2001/12/19 - Add status elements and fetch elements
  TIMAPStatusListElem = (imapMessages, imapRecent, imapUidnext, imapUidvalidity, imapUnseen);
  TIMAPStatusList = set of TIMAPStatusListElem;
  TIMAPFetchListElem = (imapFetchUid, imapFetchFlags, imapFetchInternaldate, imapFetchEnvelope, imapFetchBodystructure,
                        imapFetchBody, imapFetchBodysection, imapFetchBodypeek, imapFetchRfc822, imapFetchRfc822header,
                        imapFetchRfc822size, imapFetchRfc822text, imapFetchFast, imapFetchAll, imapFetchFull);
  TIMAPFetchList = set of TIMAPFetchListElem;
  // 2001/12/21 - Add authenticate elements - currently only none/plain implemented
  TIMAPAuthenticateListElem = (imapAuthNone,imapAuthPlain);
  TIMAPAuthenticateList = set of TIMAPAuthenticateListElem;


  TOnBeforeRetrieve = procedure( sender: TObject; const Folder: string; MsgIndex: word) of object;
  // === 2001/03/22 - Reverted Delete event back to original form for backwards compatability - James.
  TOnDeleteEvent = procedure( Sender: TObject; const Folder: string; MsgIndex: word ) of object;
  // === 2001/03/22 - Added set and clear flag event based on TIMAPFlag - James.
  // === 2001/03/22 - Added an event to watch the StoreFlags method in case somebody wants to - James.
  TOnSetFlagEvent = procedure ( Sender: TObject; const Folder: string; MsgIndex: word; flag : TIMapFlag ) of object;
  TOnClearFlagEvent = procedure ( Sender: TObject; const Folder: string; MsgIndex: word; flag : TIMapFlag ) of object;
  TOnStoreFlagsEvent = procedure ( Sender: TObject; const Folder: string; MsgIndex: word; flag : TIMapFlags ) of object;

  TToFromSocketEvent = procedure( Sender: TObject; const s: string) of object;

  // ==== 2001/02/08 - Add Checkpoint event - James.
  TOnCheckpointEvent = procedure( Sender: TObject; const Folder: string ) of object;
  // ==== 2001/02/08 - ACL Events types - James.
  // ==== 2001/03/21 - To follow the flag events above we utilize a Set and Clear event for ACL extensions - James.
  TOnSetACLEvent = procedure( Sender: TObject; const Folder: string; identifier: string; rights : TIMAPACLRights ) of object;
  TOnClearACLEvent = procedure( Sender: TObject; const Folder: string; identifier: string; rights : TIMAPACLRights ) of object;
  TOnDeleteACLEvent = procedure( Sender: TObject; const Folder: string; identifier: string ) of object;
  TOnGetACLEvent = procedure( Sender: TObject; const Folder: string ) of object;
  // ==== 2001/03/22 - Added Listrights and Myrights events to finish off ACL extension support
  TOnListrightsACLEvent = procedure( Sender: TObject; const Folder: string; identifier: string ) of object;
  TOnMyrightsACLEvent = procedure( Sender: TObject; const Folder: string ) of object;
  // ==== 2001/12/18 - Add Noop and other events - James.
  TOnNoopEvent = procedure( Sender: TObject; const Folder: string; NumMsgs: word; RecentMsgs: word ) of object;
  TOnStatusEvent = procedure( Sender: TObject; const Folder: string; elements: TIMAPStatusList ) of object;
  TOnFetchEvent = procedure( Sender: TObject; const Folder: string; const messageset: string; elements: TIMAPFetchList ) of object;
  TOnFetchBodyEvent = procedure( Sender: TObject; const Folder: string; const messageset: string; elements: TIMAPFetchList;
                                 const sectionstring : string; const partialstring : string ) of object;
  TOnIdleEvent = procedure( Sender: TObject; const Folder: string; NumMsgs: word; RecentMsgs: word; idleDone: boolean ) of object;
  TOnStoreFlagsUIDEvent = procedure ( Sender: TObject; const Folder: string; UID: string; flag : TIMapFlags ) of object;
  TOnFetchUIDEvent = procedure( Sender: TObject; const Folder: string; const UIDset: string; elements: TIMAPFetchList ) of object;
  TOnSearchEvent = procedure( Sender: TObject; const Folder: string; const charsetstr: string; searchcriteria: string ) of object;
  TOnSearchUIDEvent = procedure( Sender: TObject; const Folder: string; const charsetstr: string; searchcriteria: string ) of object;
  TOnCopyEvent = procedure( Sender: TObject; const MessageSet : string; const FolderName : string ) of object;
  TOnCopyUIDEvent = procedure( Sender: TObject; const UIDset : string; const FolderName : string ) of object;
  // === 2002/01/02 - Missed a couple - adding for completeness
  TOnListEvent = procedure( Sender: TObject; const referencename: string; const mailboxwildcard: string ) of object;

// ================ SakIMAP ================

  TSakIMAP = class(TComponent)
  private
    { Private declarations }
    FSocket: TClientSocket;

    FUserId: string;
    FUserPasswd: string;
    FPort: string;
    FHost: string;
    FIMAPError: boolean;
    FIMAPErrorString: string;
    FStrErrorInFormatOfMsg: string;
    FConnected: boolean;
    FReceivingMsg: boolean;
    FReplyCode: string;
    FReplyString: string;
    FCanceled: boolean;
    FValidated: boolean;
    FACLExtensionsOn: boolean;         // ==== 2001/02/08 - Turn ACL List extensions on - James ====
    FAutoMarkRead: boolean;            // ==== 2001/02/08 - Automatically Mark messages read on download - James ====
    FOperationTimeout : LongInt;       // ==== 2001/12/10 - Timeout value for send/receive operations - in ms - James ===
    FCloseOnError : boolean;           // ==== 2001/12/20 - Close on errors - on by default, but can be disabled - James ===
    FIdleDoneRequested : boolean;      // ==== 2001/12/20 - Interrupt the Idle - James.
    FDefaultListWildcard : string;     // ==== 2002/01/02 - Default List wildcard - * or %
    FListFullHierarchy : boolean;      // ==== 2002/01/02 - List uses * by default if true, % if false

    imap_tag: string;
    FSelectedFolder: string;
    FDelim: char;
    FFolderList: TStringList;
    FSUBFolderList: TStringList;       // === 2001/02/09 - SubFolderList - James.
    FAvFolderList: TStringList;        // === 2001/12/11 - AvFolderList - James.
    FAvSUBFolderList: TStringList;     // === 2001/12/11 - AvSubFolderList - James.
    FTempFolderList: TStringList;      // === 2001/12/12 - TempFolderList - James.
    FTempSUBFolderList: TStringList;   // === 2001/12/12 - TempSUBFolderList - James.

    FMsgsCount: word;
    FNewMsgsCount: word;
    FRecentMsgsCount: word;            // === 2001/12/18 - Track recent messages too.

    FGetUIDLsOnRetrieve: boolean; // Added by Peter Honan
    FGetIMAPFlagsOnRetrieve: boolean; // Added by Peter Honan

    FOnConnect: TNotifyEvent;
    FOnDisconnect: TNotifyEvent;
    FOnError: TOnError;
    FOnBeforeRetrieve: TOnBeforeRetrieve;
    FOnAfterRetrieve: TNotifyEvent;
    FOnDelete: TOnDeleteEvent;
    FOnLookup: TSocketNotifyEvent;
    FOnConnecting: TSocketNotifyEvent;
    FOnSendTextToSocket: TToFromSocketEvent;
    FOnReceiveTextFromSocket: TToFromSocketEvent;
  // ==== 2001/02/07 - New Event Support for marking messages - James.
  // ==== 2001/03/22 - Updated to conform to newer type definition - James
    FOnSetFlag: TOnSetFlagEvent;
    FOnClearFlag : TOnClearFlagEvent;
    FOnStoreFlags : TOnStoreFlagsEvent;
  // ==== 2001/03/22 - Call it CheckPointEvent now.... - James.
    FOnCheckpoint : TOnCheckpointEvent;
  // === 2001/02/08 - New Event Support for ACL lists - James.
  // === 2001/03/22 - Added support for clear ACL event ( removing Boolean flags ) - James.
    FOnSetACL : TOnSetACLEvent;
    FOnClearACL : TOnClearACLEvent;
    FOnDeleteACL : TOnDeleteACLEvent;
    FOnGetACL : TOnGetACLEvent;
    FOnListrightsACL : TOnListrightsACLEvent;
    FOnMyrightsACL : TOnMyrightsACLEvent;
   // ==== 2001/12/18 - Attempt to round out IMAP4rev1 command capabilities ===
    FOnCapability: TNotifyEvent;
    FOnNoop : TOnNoopEvent;
    FOnStatus : TOnStatusEvent;
    FOnFetch : TOnFetchEvent;
    FOnFetchBody : TOnFetchBodyEvent;
    FOnIdle : TOnIdleEvent;
    FOnStoreFlagsUID : TOnStoreFlagsUIDEvent;
    FOnFetchUID : TOnFetchUIDEvent;
    FOnSearch : TOnSearchEvent;
    FOnSearchUID : TOnSearchUIDEvent;
    FOnCopy : TOnCopyEvent;
    FOnCopyUID : TOnCopyUIDEvent;
    FOnNamespace : TNotifyEvent;
    FOnList : TOnListEvent;

    procedure FDoConnect( Sender: TObject; Socket: TCustomWinSocket);
    procedure FDoDisconnect( Sender: TObject; Socket: TCustomWinSocket);
    procedure FDoErr;
    procedure FDoError( Sender: TObject; Socket: TCustomWinSocket;
                        ErrorEvent: TErrorEvent; var ErrorCode: Integer);
    procedure FDoDelete( const MsgIndex: word ); // == 2001/03/22 - Reverted to keep backward compatability - James
    procedure FDoBeforeRetrieve( const MsgIndex: word);
    procedure FDoAfterRetrieve;
    procedure FSendTextToSocket( const txt: string);
    procedure FDoLookup( Sender: TObject; Socket: TCustomWinSocket);
    procedure FDoConnecting( Sender: TObject; Socket: TCustomWinSocket);
  // ==== 2001/12/10 - Possible infinite loop conditions in FReceiveTextFromSocket and
  // ====    FReceiveMailFromSocket found and mitigated with a timeout value ( user defined ).
    procedure FReceiveTextFromSocket;
    procedure FReceiveMailFromSocket( RawMail: TStringList);
  // ==== 2001/02/07 - New Event Method for marking messages - James.
  // ==== 2001/03/22 - Replaced with more compact method - James.
    procedure FDoSetFlag( const MsgIndex: word; setFlag : TIMAPFlag );
    procedure FDoClearFlag( const MsgIndex: word; setFlag : TIMAPFlag );
  // ==== 2001/03/22 - Deal with Store flags - James.
    procedure FDoStoreFlags( const MsgIndex: word; Flags: TIMAPFlags );
  // ==== 2001/02/08 - Checkpoint event - James.
    procedure FDoCheckpoint();
  // ==== 2001/02/08 - Support for ACL Extenstions - James.
  // ==== 2001/03/22 - Updated to conform to enum standard - James.
    procedure FDoSetACL( const folder: string; const identifier: string; rights : TIMAPACLRights );
    procedure FDoClearACL( const folder: string; const identifier: string; rights : TIMAPACLRights );
    procedure FDoDeleteACL( const folder: string; const identifier : string );
    procedure FDoGetACL( const folder: string );
    procedure FDoListrightsACL( const folder : string; const identifier : string );
    procedure FDoMyrightsACL( const folder : string );
   // ==== 2001/12/18 - Attempt to round out IMAP4rev1 command capabilities ===
    procedure FDoCapability();
    procedure FDoNoop( const folder : string; const numMsgs: word; const recentMsgs: word );
    procedure FDoStatus( const folder : string; elements: TIMAPStatusList );
    procedure FDoFetch( const folder : string; const messageset: string; elements: TIMAPFetchList );
    procedure FDoFetchBody( const folder : string; const messageset: string; elements: TIMAPFetchList;
                            const sectionstring : string; const partialstring : string );
    procedure FDoIdle( const folder : string; const numMsgs: word; const recentMsgs: word;
                       const idleDone: boolean );
    procedure FDoStoreFlagsUID( const UIDstr: string; Flags: TIMAPFlags );
    procedure FDoFetchUID( const folder : string; const UIDset: string; elements: TIMAPFetchList );
    procedure FDoSearch( const folder : string; const charsetstr: string; const searchcriteria: string );
    procedure FDoSearchUID( const folder : string; const charsetstr: string; const searchcriteria: string );
    procedure FDoCopy( const MessageSet: string; const FolderName: string );
    procedure FDoCopyUID( const UIDset: string; const FolderName: string );
   // ==== 2002/01/02 - Finish up command set
    procedure FDoNamespace();
    procedure FDoList( const referencename : string; const mailboxwildcard : string );
    procedure SetFullHierarchy( const setval : boolean );           // === 2002/01/02 - Set List token
    function  RetrieveMessageExt( const MsgIndex: word;
                                  const OnlyHeaders: boolean;
                                  var SakMsg: TSakMsg): boolean;
    function  RetrieveRetrieveMessageRangeExt( Start, Finish: integer;
                                               var SakMsgList: TSakMsgList;
                                               const OnlyHeaders: boolean): integer;
    function  RetrieveAllMessagesExt( var SakMsgList: TSakMsgList;
                                      const OnlyHeaders: boolean): integer;
    function  SearchRetrieveMessagesExt(Criteria: string;
                                        var SakMsgList: TSakMsgList;
                                        const OnlyHeaders: boolean): integer;
    function  ImapGetNextTag(const s: string): string;
    function  GetHierarchyDelim: char;
    function  GetFolderList: TStringList;
  // ==== 2001/02/09 - Get Subscribed Folder List - James.
    function  GetSUBFolderList: TStringList;    // ==== 2001/02/09 - Get Subscribed Folder List - James.
    function  DoForceAbend : boolean;           // ==== 2001/12/10 - Force the Connection To Abend - James.
    function  GetAvFolderList: TStringList;     // ==== 2001/12/11 - Get Available Folder List - James
    function  GetAvSUBFolderList: TStringList;  // ==== 2001/02/09 - Get Available Subscribed Folder List - James.
    function  LowLevelFetch( const messageset: string; elements: TIMAPFetchList; const sendonly : boolean;
                             const UIDfetch: boolean ) : boolean;   // === 2001/12/21 - UID and/or only send ===
    function  LowLevelSearch( const charsetstr: string; const searchcriteria: string;
                              const UIDsearch: boolean ): boolean;   // === 2001/12/21 - UID or msgno
    function  LowLevelCopy( const MessageSet : string; const FolderName : string;
                            const UIDcopy: boolean ): boolean;     // === 2001/12/21 - UID or msgno
    function  LowLevelStoreFlags( const MsgIndex: string; Flags: TIMAPFlags;
                                  const UIDstore: boolean ): boolean;
    function  FReceiveTextFromSocketUntilToken( const token: string ) : boolean; // === Use token instead of tag ===
  protected
    { Protected declarations }
  public
    { Public declarations }
    constructor Create(AOwner: TComponent); override;
    destructor  Destroy; override;
    function  Connect: boolean;
    function  Disconnect: boolean;              // === 2001/02/07 - Return Status as well - James.
    function  Login: boolean;
    function  SelectFolder(const FolderName: string): boolean;
    function  CreateFolder(const FolderName: string): boolean;
    function  DeleteFolder(const FolderName: string): boolean;
    function  DeleteMessage( const MsgIndex: word ) : boolean;   // === 2001/03/22 - Revert to backward compatability - James.
    function  UndeleteMessage( const MsgIndex: word ) : boolean; // === 2001/03/22 - Symmetric call for convenience - James.
    function  Expunge: boolean;
    function  AppendMessage( const Folder: string; const RawMail: TStringList): boolean;
    procedure Cancel;
    function  RetrieveMessage( const MsgIndex: word; var SakMsg: TSakMsg): boolean;
    function  RetrieveMessageByUID( const UID: string; var SakMsg: TSakMsg): boolean;
    function  RetrieveMessageOnlyHeaders( const MsgIndex: word; var SakMsg: TSakMsg): boolean;
    function  RetrieveMessageRange( Start, Finish: integer;  var SakMsgList: TSakMsgList): integer;
    function  RetrieveMessageRangeOnlyHeaders( Start, Finish: integer; var SakMsgList: TSakMsgList): integer;
    function  RetrieveAllMessages( var SakMsgList: TSakMsgList): integer;
    function  RetrieveAllMessagesOnlyHeaders( var SakMsgList: TSakMsgList): integer;
    function  StoreFlags( const MsgIndex: word; Flags: TIMAPFlags): boolean;
    function  FetchMsgIndex( const UID: string): word;
    function  SearchRetrieveMessages(Criteria: string; var SakMsgList: TSakMsgList): integer;
    function  SearchRetrieveMessageHeaders(Criteria: string; var SakMsgList: TSakMsgList): integer;
  // ==== 2001/02/07 - New Method Support for marking message flags - James.
  // ==== 2001/03/22 - Methods redesigned to fit desired coding conventions - James.
    function  SetMessageFlag( const MsgIndex: word; setFlag : TIMAPFlag ) : boolean;
    function  ClearMessageFlag( const MsgIndex: word; setFlag : TIMAPFlag ) : boolean;
  // ==== 2001/02/08 - Checkpoint method - James.
    function  DoCheckpoint() : boolean;
  // ==== 2001/02/08 - Support for ACL Extenstions - James.
  // ==== 2001/03/22 - Updated to conform to cleaner model - James.
    function  SetACL( const folder: string; const identifier: string; rights : TIMAPACLRights ) : boolean;
    function  ClearACL( const folder: string; const identifier: string; rights : TIMAPACLRights ) : boolean;
    function  DeleteACL( const folder: string; const identifier: string ) : boolean;
    function  GetACL( const folder: string ) : boolean;
    function  ListrightsACL( const folder: string; const identifier: string ) : boolean;
    function  MyrightsACL( const folder: string; const identifier: string ) : boolean;
  // 2001/02/09 - Rename folder method added - James.
    function  RenameFolder(const OriginalFolderName: string; const NewFolderName : string ): boolean;
  // 2001/02/09 - Subscribe, Unsubscribe and LSub methods added.  Lsub is very primitive, requiring the caller to
  //              examine the results string to extract the subscription lists manually.  Later a
  //              TStringList parsed result may be made available - James.
    function  SubscribeFolder(const FolderName : string ) : boolean;
    function  UnsubscribeFolder(const FolderName : string ) : boolean;
    function  LSUBFolder(const ReferenceName : string; const WildcardMailbox : string ) : boolean;
  // 2001/02/09 - Copy method added. This requires that a mailbox be currently selected
  //              as well as a message set and valid folder destination - James.
    function  CopyMessageToFolder(const MessageSet : string; const FolderName : string ) : boolean;
  // ==== 2001/12/18 - Attempt to round out IMAP4rev1 command capabilities ===
    function  Capability() : boolean;
    function  Noop() : boolean;
    function  Status( const folder: string; elements: TIMAPStatusList ) : boolean;
    function  Fetch( const messageset: string; elements: TIMAPFetchList ) : boolean;
    function  FetchBody( const messageset: string; const peekonly: boolean; const sectionstring: string; const partialstring: string ) : boolean;
    function  ExamineFolder(const FolderName: string): boolean;                                // === 2001/12/20 - Readonly Select
    function  CloseSelectedFolder(): boolean;
    function  Idle( timeout: word ): boolean;
    function  Search( const charsetstr: string; const searchcriteria: string ): boolean;       // == 2001/12/21 - Search results are msgno
    function  UIDSearch( const charsetstr: string; const searchcriteria: string ): boolean;    // == 2001/12/21 - Search results are UID
    function  UIDStoreFlags( const UIDstr: string; Flags: TIMAPFlags): boolean;                // == 2001/12/21 - Store by UID
    function  UIDFetch( const UIDset: string; elements: TIMAPFetchList ) : boolean;            // == 2001/12/21 - Fetch by UID
    function  UIDCopyMessageToFolder(const UIDSet : string; const FolderName : string ) : boolean;  // == 2001/12/21 - Copy by UID
    function  Authenticate( authtype: TIMAPAuthenticateList; const paramList : TStringList ) : boolean; // === 2001/12/21 - Plain only
    function  Namespace() : boolean; // === 2002/01/02 - Get namespace information ( if available )
    function  List( const referencename : string; const mailboxwildcard : string ) : boolean;  // === 2002/01/02 - List command

    property  IMAPError: boolean read FIMAPError;
    property  ErrorString: string read FIMAPErrorString;
    property  MsgsCount: word read FMsgsCount;
    property  NewMsgsCount: word read FNewMsgsCount;
    property  RecentMsgsCount: word read FRecentMsgsCount;                      // === 2001/12/18 - Track recent messages too.
    property  Connected: boolean read FConnected;
    property  ReplyCode: string read FReplyCode;
    property  ReplyString: string read FReplyString;
    property  Canceled: boolean read FCanceled;
    property  HierarchyDelim: char read GetHierarchyDelim;
    property  FolderList: TStringList read GetFolderList;
    property  Validated: boolean read FValidated;
    property  ACLExtensionsOn : boolean read FACLExtensionsOn write FACLExtensionsOn; // === 2001/02/08 - Turn On ACL support - James
    property  AutoMarkRead : boolean read FAutoMarkRead write FAutoMarkRead;          // ==== 2001/02/08 - Automatically Mark messages read on download - James
    property  SUBFolderList: TStringList read GetSUBFolderList;                 // ==== 2001/02/09 - Get Subscribed Folder List - James.
    property  ForceAbend : boolean read DoForceAbend;                           // ==== 2001/12/10 - Force the Connection To Abend - James.
    property  AvFolderList: TStringList read GetAvFolderList;                   // === 2001/12/11 - Get Available (exist) Folder List - James.
    property  AvSUBFolderList: TStringList read GetAvSUBFolderList;             // === 2001/12/11 - Get Available (exist) subscribed Folder List - James.
    property  IdleDoneRequested : boolean write FIdleDoneRequested;             // ==== 2001/12/20 - Interrupt the Idle - James.
  published
    { Published declarations }
    property  UserId: string read FUserId write FUserId;
    property  UserPasswd: string read FUserPasswd write FUserPasswd;
    property  Port: string read FPort write FPort;
    property  Host: string read FHost write FHost;
    property  StrErrorInFormatOfMsg: string read FStrErrorInFormatOfMsg write FStrErrorInFormatOfMsg;
    property  SelectedFolder: string read FSelectedFolder;
    property  GetUIDLsOnRetrieve: boolean read FGetUIDLsOnRetrieve write FGetUIDLsOnRetrieve;
    property  GetIMAPFlagsOnRetrieve: boolean read FGetIMAPFlagsOnRetrieve write FGetIMAPFlagsOnRetrieve;
    property  OnConnect: TNotifyEvent read FOnConnect write FOnConnect;
    property  OnDisconnect: TNotifyEvent read FOnDisconnect write FOnDisconnect;
    property  OnError: TOnError read FOnError write FOnError;
    property  OnBeforeRetrieve: TOnBeforeRetrieve read FOnBeforeRetrieve write FOnBeforeRetrieve;
    property  OnAfterRetrieve: TNotifyEvent read FOnAfterRetrieve write FOnAfterRetrieve;
    property  OnDelete: TOnDeleteEvent read FOnDelete write FOnDelete;
    property  OnLookup: TSocketNotifyEvent read FOnLookup write FOnLookup;
    property  OnConnecting: TSocketNotifyEvent read FOnConnecting write FOnConnecting;
    property  OnSendTextToSocket: TToFromSocketEvent read FOnSendTextToSocket write FOnSendTextToSocket;
    property  OnReceiveTextFromSocket: TToFromSocketEvent read FOnReceiveTextFromSocket write FOnReceiveTextFromSocket;
   // === 2001/02/07 - New Method Support for marking message flags - James.
   // === 2001/03/22 - Updated to reflect the use of 2 events with enum type - James.
    property  OnSetFlag : TOnSetFlagEvent read FOnSetFlag write FOnSetFlag;
    property  OnClearFlag : TOnClearFlagEvent read FOnClearFlag write FOnClearFlag;
    property  OnStoreFlags : TOnStoreFlagsEvent read FOnStoreFlags write FOnStoreFlags;
    property  OnCheckpoint : TOnCheckpointEvent read FOnCheckpoint write FOnCheckpoint;
  // ==== 2001/02/08 - Support for ACL Extenstions - James.
  // ==== 2001/03/22 - Updated to conform to coding standard.
    property  OnSetACL : TOnSetACLEvent read FOnSetACL write FOnSetACL;
    property  OnDeleteACL : TOnDeleteACLEvent read FOnDeleteACL write FOnDeleteACL;
    property  OnGetACL : TOnGetACLEvent read FOnGetACL write FOnGetACL;
    property  OnListrightsACL : TOnListrightsACLEvent read FOnListrightsACL write FOnListrightsACL;
    property  OnMyrightsACL : TOnMyrightsACLEvent read FOnMyrightsACL write FOnMyrightsACL;
    property  OperationTimeout : LongInt read FOperationTimeout write FOperationTimeout;  // === 2001/12/10 - Timeout value for send/receive operations - in ms ===
    property  CloseOnError : boolean read FCloseOnError write FCloseOnerror;              // ==== 2001/12/20 - Close on errors - on by default, but can be disabled - James ===
    property  OnCapability : TNotifyEvent read FOnCapability write FOnCapability;
    property  OnNoop : TOnNoopEvent read FOnNoop write FOnNoop;
    property  OnStatus : TOnStatusEvent read FOnStatus write FOnStatus;
    property  OnFetch : TOnFetchEvent read FOnFetch write FOnFetch;
    property  OnFetchBody : TOnFetchBodyEvent read FOnFetchBody write FOnFetchBody;
    property  OnIdle : TOnIdleEvent read FOnIdle write FOnIdle;
    property  OnStoreFlagsUID : TOnStoreFlagsUIDEvent read FOnStoreFlagsUID write FOnStoreFlagsUID;
    property  OnFetchUID : TOnFetchUIDEvent read FOnFetchUID write FOnFetchUID;
    property  OnSearch : TOnSearchEvent read FOnSearch write FOnSearch;
    property  OnSearchUID : TOnSearchUIDEvent read FOnSearchUID write FOnSearchUID;
    property  OnCopy : TOnCopyEvent read FOnCopy write FOnCopy;
    property  OnCopyUID : TOnCopyUIDEvent read FOnCopyUID write FOnCopyUID;
    property  OnNamespace : TNotifyEvent read FOnNamespace write FOnNamespace;
    property  OnList : TOnListEvent read FOnList write FOnList;
    property  ListFullHierarchy : boolean read FListFullHierarchy write SetFullHierarchy;
end;

implementation

uses Forms, windows, SakMime;

// === 2001/12/28 - Allow use of TWinsocketStream for D3 or higher.  There is
// ===    some indication it may not exist on some versions of D3 ... if so,
// ===    comment out the immediate {$IFDEF}..{$ENDIF} sections below. - James.
// === 2001/12/28 - To avoid use of TWinsocketStream simply comment out the
// ===    lines {$IFDEF}..{$ENDIF} inclusive ( in case of an unforseen backward
// ===    compatability issue.
{$IFDEF VER100}
   {$DEFINE SAK_WSOCK_CANUSESOCKSTREAM}
{$ENDIF}
{$IFDEF VER120}
   {$DEFINE SAK_WSOCK_CANUSESOCKSTREAM}
{$ENDIF}
{$IFDEF VER125}
   {$DEFINE SAK_WSOCK_CANUSESOCKSTREAM}
{$ENDIF}
{$IFDEF VER130}
   {$DEFINE SAK_WSOCK_CANUSESOCKSTREAM}
{$ENDIF}
{$IFDEF VER140}
   {$DEFINE SAK_WSOCK_CANUSESOCKSTREAM}
{$ENDIF}

const
  crlf = #13#10;
  IMAP_BUFFER_SIZE = 2048;
{$IFDEF SAK_WSOCK_CANUSESOCKSTREAM}
  IMAP_MAXSOCKSTREAMWAIT = 600;
{$ENDIF}


// ==== 2001/03/22 - Utilize SakMsg TIMAPACLRight as the indexing for
// ====              a constant array of ACL rights characters - James.
const TIMAPNumACLRights = 9;
const TIMAPACLRightsChar : array[0..TIMAPNumACLRights-1] of Char =
      ( 'l', 'r', 's', 'w', 'i', 'p', 'c', 'd', 'a' );
// ==== 2001/12/19 - Status list elements and fetch list elements.
const TIMAPNumStatusElems = 5;
const TIMAPStatusListStrings : array[0..TIMAPNumStatusElems-1] of string =
      ( 'MESSAGES', 'RECENT', 'UIDNEXT', 'UIDVALIDITY', 'UNSEEN' );
const TIMAPNumFetchElems = 15;
const TIMAPFetchListStrings : array[0..TIMAPNumFetchElems-1] of string =
      ( 'UID', 'FLAGS', 'INTERNALDATE', 'ENVELOPE', 'BODYSTRUCTURE', 'BODY', 'BODY', 'BODY.PEEK', 'RFC822',
        'RFC822.HEADER', 'RFC822.SIZE', 'RFC822.TEXT', 'FAST', 'ALL', 'FULL' );

// ==== 2001/12/10 - FixFolderName - Looks for errors in the folder name
// ====     structure as it stands and returns a proper form.
// ====     Currently if spaces are present, returns a "quoted-string" version.
function FixFolderName( FolderName : string ) : string;
const
   SpaceStr : string = ' ';
   QuotesStr : string = '"';
begin
    result := FolderName;
    // ==== 2002/01/03 - Avoid recursive folder fixes with extra ""'s or
    // ====    fixing a folder name the client has already quoted for us - James.
    if ( Pos ( SpaceStr, FolderName ) > 0 )
       and ( Pos( QuotesStr, FolderName ) = 0 ) then
    begin
       result := '"' + FolderName + '"';
    end;
end;


{ ***************** TSakIMAP ***************** }


constructor TSakIMAP.Create( AOwner: TComponent);
begin
  inherited create( AOwner);
  FGetUIDLsOnRetrieve := True;
  FGetIMAPFlagsOnRetrieve := True;
  FConnected := False;
  FValidated := False;
  FPort := '143';
  FReceivingMsg := False;
  imap_tag := '';
  FFolderList := TStringList.Create;
  FStrErrorInFormatOfMsg := 'Error en el formato del mensaje.';
  FSocket := nil;
  FACLExtensionsOn := False;                 // ==== 2001/02/08 - ACL extension off by default - James.
  FAutoMarkRead := True;                     // ==== 2001/02/08 - Automatically Mark messages read on download - James.
  FSUBFolderList := TStringList.Create;      // === 2001/02/09 - FSUBFolderList - must create and free - James.
  FAvFolderList := TStringList.Create;       // === 2001/12/11 - AvFolderList - James.
  FAvSUBFolderList := TStringList.Create;    // === 2001/12/11 - AvSubFolderList - James.
  FTempFolderList := TStringList.Create;     // === 2001/12/11 - TempFolderList - James.
  FTempSUBFolderList := TStringList.Create;  // === 2001/12/12 - TempSUBFolderList - James.
  FSelectedFolder := '';
  FMsgsCount := 0;
  FNewMsgsCount := 0;
  FRecentMsgsCount := 0;
  FCloseOnError := True;                     // ==== 2001/12/20 - Close on errors - James ===
  FIdleDoneRequested := False;               // ==== 2001/12/20 - Interrupt the Idle - James.
  FDefaultListWildcard := '"*"';             // ==== 2002/01/02 - Default List wildcard - *
  FListFullHierarchy := True;                // ==== 2002/01/02 - List uses * by default
  FDelim := #0;                              // ==== 2002/01/02 - For completeness

  if ( FOperationTimeout = 0 ) then
  begin
     FOperationTimeout :=  30000; // === 2001/12/10 - Timeout value for send/receive operations ===
  end;
  if not ( csDesigning in ComponentState) then
  begin
    FSocket := TClientSocket.create( AOwner);
    FSocket.ClientType := ctBlocking;
//    FSocket.OnRead := FSocketDoRead;
    FSocket.OnLookup := FDoLookup;
    FSocket.OnConnecting := FDoConnecting;
    FSocket.OnConnect := FDoConnect;
    FSocket.OnDisconnect := FDoDisconnect;
    FSocket.OnError := FDoError;
  end;
end;

destructor TSakIMAP.Destroy;
begin
//  if FSocket <> nil then
//    FreeAndNil(FSocket);
  FFolderList.Free;
  FSUBFolderList.Free();        // === 2001/02/09 - FSUBFolderList - must free - James.
  FSUBFolderList := nil;
  FAvFolderList.Free();         // === 2001/12/11 - AvFolderList - must free -James.
  FAvFolderList := nil;
  FAvSUBFolderList.Free();      // === 2001/12/11 - AvSubFolderList - must free - James.
  FAvSUBFolderList := nil;
  FTempFolderList.Free();       // === 2001/12/12 - TempFolderList - James.
  FTempFolderList := nil;
  FTempSUBFolderList.Free();    // === 2001/12/12 - TempSUBFolderList - James.
  FTempSUBFolderList := nil;

  inherited Destroy;
end;

function TSakIMAP.Connect: boolean;
var
  buff: array[1..IMAP_BUFFER_SIZE] of char;
  x: integer;
{$IFDEF SAK_WSOCK_CANUSESOCKSTREAM}
  startTickCount : LongInt;     // === 2001/12/10 - Track Start of operation
  currentTickCOunt : LongInt;   // === 2001/12/10 - Track Current ticks
  tempsockstream : TWinSocketStream;
{$ENDIF}
begin
  result := False;

  // === 2001/02/08 - Connect will *hang* if already connected because
  // ===    the server is not ready to receive data.  Protect against
  // ===    this with a true return if already connected - James.
  if ( FConnected ) then
  begin
    result := True;
    exit;
  end;

  FMsgsCount := 0;
  FSocket.port := StrToIntDef( FPort, 110);
  if sak_IsIPAddress(FHost) then
  begin
    FSocket.Address := FHost;        // Roger
    FSocket.Host := '';              // Roger
  end else
  begin
    FSocket.Host := FHost;          // Sergio
    FSocket.Address := '';          // Sergio
  end;

  FIMAPError := False;
  try
    FSocket.Open;
  except
    FIMAPError := True;
    exit;
  end;

  if FConnected then
  begin
    result := True;
    Application.ProcessMessages;

    {$IFDEF SAK_WSOCK_CANUSESOCKSTREAM}
    tempsockstream := nil;
    try
       tempsockstream := TWinSocketStream.Create( FSocket.Socket, IMAP_MAXSOCKSTREAMWAIT );
       startTickCount := GetTickCount();
    {$ENDIF}

    {$IFDEF SAK_WSOCK_CANUSESOCKSTREAM}
       while ( tempsockstream.WaitForData(IMAP_MAXSOCKSTREAMWAIT) = False ) do
       begin
          currentTickCount := GetTickCount();
          if ( FSocket.Socket.Connected = False ) then
             raise Exception.create( 'Receive Failed.  No Longer Connected.' );
          if ( currentTickCount - startTickCount > FOperationTimeout ) then
          begin
             raise Exception.create( 'Receive Timeout.  No Data Available.' );
          end;
          Application.ProcessMessages;
          Sleep(0);
       end;
       x := FSocket.Socket.ReceiveBuf( buff, IMAP_BUFFER_SIZE);
    {$ELSE}
       x := FSocket.Socket.ReceiveBuf( buff, IMAP_BUFFER_SIZE);
    {$ENDIF}

    FReplyString := copy( buff, 1, x);
    //FReceiveTextFromSocket;
    FCanceled := False;

    {$IFDEF SAK_WSOCK_CANUSESOCKSTREAM}
    finally
       if assigned( tempsockstream ) then
          tempsockstream.Free();
       // tempsockstream := nil;
    end;
    {$ENDIF}
  end;
  FSelectedFolder := '';
  FMsgsCount := 0;
  FNewMsgsCount := 0;
  FRecentMsgsCount := 0;
  FDelim := #0;    // === 2002/01/02 - Must clear hierarchy delimeter on Connection.
                   // ===    in case we are connecting to multiple servers.
end;

// === 2001/03/20 - Just desired to make the return for Disconnect
// ===              Symmetric with connect - James.
function TSakIMAP.Disconnect : boolean;
var
  cmd: string;
begin
  result := False;
  if FConnected then
  begin
    imap_tag := ImapGetNextTag(imap_tag);
    cmd := imap_tag + ' LOGOUT' + crlf;
    FSendTextToSocket( cmd);
    FReceiveTextFromSocket;
    if FIMAPError then
    begin
      FDoErr;
      if ( FCloseOnError = False ) then // === 2001/12/20 - Force close - James ===
      begin
         FSocket.close;
         FConnected := False;
         FValidated := False;
         result := False;
         Application.ProcessMessages;
      end;
    end else
    begin
      FSocket.close;
      FConnected := False;
      FValidated := False;
      result := True;
      Application.ProcessMessages;
    end;
  end
  else
  begin
    result := True;
  end;
end;

procedure TSakIMAP.Cancel;
begin
  FCanceled := True;
end;

function TSakIMAP.Login: boolean;
var
  cmd: string;
begin
  result := False;
  if not FConnected then
  begin
    raise Exception.create( 'Invalid Op. Not connected.');
  end;

  imap_tag := ImapGetNextTag(imap_tag);
  cmd := imap_tag + ' LOGIN ' + UserId + ' ' + UserPasswd + crlf;
  FSendTextToSocket( cmd);
  FReceiveTextFromSocket;

  if FIMAPError then
  begin
    FDoErr;
    exit;
  end;

  result := pos( imap_tag + ' OK', UpperCase( FReplyString)) > 0;
  FValidated := result;
  // you can inspect ReplyString to see what happened in case
  // of returning False
end;

// GetHierarchyDelim gets the char, that is used by the IMAP server to separate
// folder names like \ in DOS path
// Peter Nagel
function TSakIMAP.GetHierarchyDelim: char;
var
  po: integer;
begin
  if not FConnected then
    raise Exception.create( sak_esNotConnected);
  if FDelim = #0 then
  begin   //only fetch if not already set
    if ( List( '""', '""' ) = True ) then   // === 2002/01/02 - Use LIST command
    begin
       po := Pos('"', FReplyString);
       if (length(FReplyString) > po) and not FIMAPError then
         FDelim := FReplyString[po+1]
       else
         FDelim := #0;
    end;
  end;
  result := FDelim;
end;

function TSakIMAP.SelectFolder(const FolderName: string): boolean;
const
   Nattrstart : string = '(';
   Nattrend   : string = ')';
   NoMessagesstr : string = 'MESSAGES ';
   NoRecentstr : string = 'RECENT ';
   NoUnseenstr : string = 'UNSEEN ';
var
  cmd, tmpFolder: string;
  ponastart, ponaend, currpos : integer;
  nattrstring, currstring : string;
  statusresult : boolean;
begin
  //result := False;
  tmpFolder := trim(FolderName);
  tmpFolder := FixFolderName( tmpFolder );

  // === 2002/01/03 - Holger Mauermann found a "client bug" in the
  // ===    behaviour of SakIMAP talking to imapd 2001a server (2001/11/26).
  // ===    The server refuses to return status information on a selected folder.
  // ===    His fix has been integrated into this changed codebase. - James.
  // === 2002/01/03 - The behaviour is now 1) Ensure any selected folder is
  // ===    closed ( to avoid the error on subsequent selects ) 2) Status the
  // ===    folder, updating MsgCount, NewMsgsCount and RecentMsgsCount
  // ===    and 3) Select the folder.  This is an extension to H. Mauermann's
  // ===    solution that allows clients who used SelectFolder to update message
  // ===    counts to still function properly with imapd 2001a - James.

  // === 2001/12/19 - Results set to 0 on failure.
  FMsgsCount := 0;
  FNewMsgsCount := 0;
  FRecentMsgsCount := 0;

  if ( FSelectedFolder <> '' ) then
  begin
     if ( CloseSelectedFolder() = False ) then
     begin
        // === 2002/01/03 - A state problem exists - client should disconnect to clear.
        //result := False;
        raise Exception.create( 'Select Failed.  Cannot close previously selected folder.' +
                                #13#10 + 'Disconnect from server if error persists.' );
        exit;
     end;
  end;

  statusresult := Status(FolderName, [imapMessages, imapUnseen, imapRecent]);

  if ( statusresult = True ) then
  begin
     ponastart := Pos( Nattrstart, FReplyString );
     ponaend := Pos( Nattrend, FReplyString );
     if ( ponastart > 0 ) and ( ponaend > 0 ) and ( ponaend > ponastart ) then
     begin
        nattrstring := copy(FReplyString, ponastart, (ponaend - ponastart) + 1);
        nattrstring := UpperCase(nattrstring);
        currpos := Pos( NoMessagesstr, nattrstring );
        if ( currpos > 0 ) then
        begin
           currstring := copy(nattrstring, currpos, (Length(nattrstring) - currpos) + 1);
           FMsgsCount := sak_StrWord2Int( currstring, 2, 0);
        end;
        currpos := Pos( NoUnseenstr, nattrstring );
        if ( currpos > 0 ) then
        begin
           currstring := copy(nattrstring, currpos, (Length(nattrstring) - currpos) + 1);
           FNewMsgsCount := sak_StrWord2Int( currstring, 2, 0);
        end;
        currpos := Pos( NoRecentstr, nattrstring );
        if ( currpos > 0 ) then
        begin
           currstring := copy(nattrstring, currpos, (Length(nattrstring) - currpos) + 1);
           FRecentMsgsCount := sak_StrWord2Int( currstring, 2, 0);
        end;
     end;
  end;

  imap_tag := ImapGetNextTag(imap_tag);
  cmd := imap_tag + ' SELECT ' + tmpFolder + crlf;

  FSendTextToSocket( cmd);
  FReceiveTextFromSocket;

  if FIMAPError then
  begin
    result := False;
    FDoErr;
    exit;
  end;

  result := pos( imap_tag + ' OK', UpperCase( FReplyString)) > 0;

  if not result then
    exit;
  // you can inspect ReplyString to see what happened in case
  // of returning False

  FSelectedFolder := trim(FolderName);   // === 2001/12/10
end;

function TSakIMAP.CreateFolder(const FolderName: string): boolean;
var
  cmd: string;
  tmpFolder : string; // === 2001/12/10
begin
  //result := False;
  imap_tag := ImapGetNextTag(imap_tag);
  tmpFolder := FixFolderName( FolderName );  // === 2001/12/10
  cmd := imap_tag + ' CREATE ' + tmpFolder + crlf;   // === 2001/12/10

  FSendTextToSocket( cmd);
  FReceiveTextFromSocket;

  if FIMAPError then
  begin
    result := False;
    FDoErr;
    exit;
  end;

  result := pos( imap_tag + ' OK', UpperCase( FReplyString)) > 0;
  // you can inspect ReplyString to see what happened in case
  // of returning False
end;

function TSakIMAP.DeleteFolder(const FolderName: string): boolean;
var
  cmd: string;
  tmpFolder : string; // === 2001/12/10
begin
  //result := False;
  imap_tag := ImapGetNextTag(imap_tag);
  tmpFolder := FixFolderName( FolderName );  // === 2001/12/10
  cmd := imap_tag + ' DELETE ' + tmpFolder + crlf;  // === 2001/12/10

  FSendTextToSocket( cmd);
  FReceiveTextFromSocket;

  if FIMAPError then
  begin
    result := False;
    FDoErr;
    exit;
  end;

  result := pos( imap_tag + ' OK', UpperCase( FReplyString)) > 0;
  // you can inspect ReplyString to see what happened in case
  // of returning False
end;

// GetFolderList fills a string list with the folder tree that is
// residing on the IMAP server
// Peter Nagel
function TSakIMAP.GetFolderList: TStringList;
var
  i, po: integer;
begin
  if not FConnected then
    raise Exception.create( sak_esNotConnected);

  if not assigned( FFolderList ) then
     FFolderList := TStringList.Create;
  FFolderList.Clear;

  if ( List( '""', FDefaultListWildcard ) = True ) then   // === 2002/01/02 - Use LIST command ===
  begin
    FFolderList.Text := FReplyString; //make lines
    FFolderlist.Delete(FFolderList.Count - 1); //delete line "imap_tag OK LIST completed"
    for i := 0 to FFolderlist.Count - 1 do
    begin
      po := Pos('"',FFolderList[i]);
      FFolderList[i] := Trim(copy(FFolderList[i], po + 3, MAXINT)); //remove stuff in front of folder name
      if Pos('"', FFolderList[i]) = 1 then  //folder name is quoted
        FFolderList[i] := copy(FFolderlist[i], 2, length(FFolderList[i]) - 2);
    end;
  end;
  result := FFolderList;
end;

// === 2001/02/09 - GetSUBFolderList - is identical to GetFolderList, but
// ===     also utilizes the public LSUBFolder method - James.
// === Full Credit to Peter Nagel for GetFolderList structure.
function TSakIMAP.GetSUBFolderList: TStringList;
var
  i, po: integer;
begin
  if not FConnected then
    raise Exception.create( sak_esNotConnected);

  if ( LSUBFolder( '""', FDefaultListWildcard ) = True ) then // ==== 2002/01/02 - Use Default List wildcard
  begin
     if not assigned( FSUBFolderList ) then
        FSUBFolderList := TStringList.Create;
     FSUBFolderList.Clear;
     FSUBFolderList.Text := FReplyString; //make lines
     FSUBFolderlist.Delete(FSUBFolderList.Count - 1); //delete line "imap_tag OK LSUB completed"
     for i := 0 to FSUBFolderlist.Count - 1 do
     begin
       po := Pos('"',FSUBFolderList[i]);
       FSUBFolderList[i] := Trim(copy(FSUBFolderList[i], po + 3, MAXINT)); //remove stuff in front of folder name
       if Pos('"', FSUBFolderList[i]) = 1 then  //folder name is quoted
         FSUBFolderList[i] := copy(FSUBFolderlist[i], 2, length(FSUBFolderList[i]) - 2);
     end;
  end;
  result := FSUBFolderList;
end;

// === 2001/12/11 - RFC2060 compliancy "issue" found.  The folderlist
// ===      that is returned *includes* unavailable folders (\noselect)
// ===      or (\nonexistent) - the latter is from an RFC IMAP extension document.
// ===   To avoid backward compatability issues we just use new folder properties
// ===      to hold the changes.
// ==== 2001/12/11 - Get Available Folder List - James
function TSakIMAP.GetAvFolderList: TStringList;
const
   Nattrstart : string = '(';
   Nattrend   : string = ')';
   Noselectstr : string = '\NOSELECT';
   Nonexiststr : string = '\NONEXISTENT';
var
  i, po: integer;
  ponastart, ponaend : integer;
  nattrstring : string;
  folderExists : boolean;
begin
  if not FConnected then
    raise Exception.create( sak_esNotConnected);

  if not assigned( FTempFolderList ) then
     FTempFolderList := TStringList.Create;
  FTempFolderList.Clear;
  if not assigned( FAvFolderList ) then
     FAvFolderList := TStringList.Create;
  FAvFolderList.Clear;

  if ( List( '""', FDefaultListWildcard ) = True ) then   // === 2002/01/02 - Use LIST command ===
  begin
    FTempFolderList.Text := FReplyString; //make lines
    FTempFolderlist.Delete(FTempFolderList.Count - 1); //delete line "imap_tag OK LIST completed"
    //ponastart := 0;
    //ponaend := 0;
    nattrstring := '';
    for i := 0 to FTempFolderList.Count - 1 do
    begin
      po := Pos('"',FTempFolderList[i]);
      ponastart := Pos( Nattrstart, FTempFolderList[i] );
      ponaend := Pos( Nattrend, FTempFolderList[i] );
      folderExists := True;
      if ( ponastart > 0 ) and ( ponaend > 0 ) and ( ponaend > ponastart ) then
      begin
         nattrstring := copy(FTempFolderList[i], ponastart, (ponaend - ponastart) + 1);
         nattrstring := UpperCase(nattrstring);
         if ( Pos( Noselectstr, nattrstring ) > 0 ) then
            folderExists := False;
         if ( Pos( Nonexiststr, nattrstring ) > 0 ) then
            folderExists := False;
      end;
      FTempFolderList[i] := Trim(copy(FTempFolderList[i], po + 3, MAXINT)); //remove stuff in front of folder name
      if Pos('"', FTempFolderList[i]) = 1 then  //folder name is quoted
        FTempFolderList[i] := copy(FTempFolderList[i], 2, length(FTempFolderList[i]) - 2);
      if ( folderExists = True ) then
        FAvFolderList.Add( FTempFolderList[i] );
    end;
  end;
  FTempFolderList.Clear;
  result := FAvFolderList;
end;

// === 2001/12/11 - RFC2060 compliancy "issue" found.  The folderlist
// ===      that is returned *includes* unavailable folders (\noselect)
// ===      or (\nonexistent) - the latter is from an RFC IMAP extension document.
// ===   To avoid backward compatability issues we just use new folder properties
// ===      to hold the changes.
// ==== 2001/02/09 - Get Available Subscribed Folder List - James.
function TSakIMAP.GetAvSUBFolderList: TStringList;
const
   Nattrstart : string = '(';
   Nattrend   : string = ')';
   Noselectstr : string = '\NOSELECT';
   Nonexiststr : string = '\NONEXISTENT';
var
  i, po: integer;
  ponastart, ponaend : integer;
  nattrstring : string;
  folderExists : boolean;
begin
  if not FConnected then
    raise Exception.create( sak_esNotConnected);

  if ( LSUBFolder( '""', FDefaultListWildcard ) = True ) then
  begin
     if not assigned( FTempSUBFolderList ) then
        FTempSUBFolderList := TStringList.Create;
     FTempSUBFolderList.Clear;
     if not assigned( FAvSUBFolderList ) then
        FAvSUBFolderList := TStringList.Create;
     FAvSUBFolderList.Clear;
     FTempSUBFolderList.Text := FReplyString; //make lines
     FTempSUBFolderList.Delete(FTempSUBFolderList.Count - 1); //delete line "imap_tag OK LSUB completed"
     //ponastart := 0;
     //ponaend := 0;
     nattrstring := '';
     for i := 0 to FTempSUBFolderList.Count - 1 do
     begin
       po := Pos('"',FTempSUBFolderList[i]);
       ponastart := Pos( Nattrstart, FTempSUBFolderList[i] );
       ponaend := Pos( Nattrend, FTempSUBFolderList[i] );
       folderExists := True;
       if ( ponastart > 0 ) and ( ponaend > 0 ) and ( ponaend > ponastart ) then
       begin
          nattrstring := copy(FTempSUBFolderList[i], ponastart, (ponaend - ponastart) + 1);
          nattrstring := UpperCase(nattrstring);
          if ( Pos( Noselectstr, nattrstring ) > 0 ) then
             folderExists := False;
          if ( Pos( Nonexiststr, nattrstring ) > 0 ) then
             folderExists := False;
       end;
       FTempSUBFolderList[i] := Trim(copy(FTempSUBFolderList[i], po + 3, MAXINT)); //remove stuff in front of folder name
       if Pos('"', FTempSUBFolderList[i]) = 1 then  //folder name is quoted
         FTempSUBFolderList[i] := copy(FTempSUBFolderList[i], 2, length(FTempSUBFolderList[i]) - 2);
       if ( folderExists = True ) then
         FAvSUBFolderList.Add( FTempSUBFolderList[i] );
     end;
  end;
  FTempSUBFolderList.Clear;
  result := FAvSUBFolderList;
end;

// === 2001/02/07 - Added +-ive or -ive flag setting ability - James.
// === 2001/03/22 - Reverted to original for backward compatability - James.
function TSakIMAP.DeleteMessage( const MsgIndex: word ): boolean;
// you must select a folder first
// and use the expunge method to really remove this messages
var
  cmd: string;
begin
  result := False;
  if not FConnected then
    raise Exception.create( sak_esNotConnected);

  imap_tag := ImapGetNextTag(imap_tag);
  cmd := imap_tag + ' STORE ' + IntToStr( MsgIndex) + ' +FLAGS (\Deleted)' + crlf;

  Application.ProcessMessages;
  if FCanceled then exit;

  FSendTextToSocket( cmd);
  FReceiveTextFromSocket;

  if FIMAPError then
  begin
    FDoErr;
    exit;
  end;

  result := pos( imap_tag + ' OK', UpperCase( FReplyString)) > 0;
  if result then
  begin
    dec( FMsgsCount);
    FDoDelete( MsgIndex );
    FDoSetFlag( MsgIndex, isDeleted );
  end;
end;

// === 2001/03/22 - Add convenient symmetric method to undelete a message - James.
function TSakIMAP.UndeleteMessage( const MsgIndex: word ): boolean;
// you must select a folder first
// and use the commit method to restore this message
begin
  result := ClearMessageFlag( MsgIndex, isDeleted );
end;


function TSakIMAP.Expunge: boolean;
// The EXPUNGE command permanently removes from the currently
// selected mailbox (folder) all messages that have the \Deleted flag set.
var
  cmd: string;
begin
  result := False;
  if not FConnected then
    raise Exception.create( sak_esNotConnected);

  imap_tag := ImapGetNextTag(imap_tag);
  cmd := imap_tag + ' EXPUNGE' + crlf;

  Application.ProcessMessages;
  if FCanceled then exit;

  FSendTextToSocket( cmd);
  FReceiveTextFromSocket;
  if FIMAPError then
  begin
    FDoErr;
    exit;
  end;

  result := pos( imap_tag + ' OK', UpperCase( FReplyString)) > 0;
end;

function TSakIMAP.AppendMessage( const Folder: string; const RawMail: TStringList): boolean;
var
  cmd: string;
  literal: string;
  x: integer;
  buff: array[1..IMAP_BUFFER_SIZE] of char;
  tmpFolder : string; // === 2001/12/10
{$IFDEF SAK_WSOCK_CANUSESOCKSTREAM}
  startTickCount : LongInt;     // === 2001/12/10 - Track Start of operation
  currentTickCOunt : LongInt;   // === 2001/12/10 - Track Current ticks
  tempsockstream : TWinSocketStream;
{$ENDIF}
begin
  result := False;
  if not FConnected then
    raise Exception.create( sak_esNotConnected);

  Application.ProcessMessages;
  if FCanceled then exit;

  literal := IntToStr( length(RawMail.Text));

  imap_tag := ImapGetNextTag(imap_tag);
  tmpFolder := FixFolderName( Folder );  // === 2001/12/10
  cmd := imap_tag + ' APPEND ' + tmpFolder + ' {' + literal + '}' + crlf;  // == 2001/12/10

  FSendTextToSocket( cmd);
  Application.ProcessMessages;

  x := 0;
  {$IFDEF SAK_WSOCK_CANUSESOCKSTREAM}
  tempsockstream := nil;
  try
     tempsockstream := TWinSocketStream.Create( FSocket.Socket, IMAP_MAXSOCKSTREAMWAIT );
     startTickCount := GetTickCount();
  {$ENDIF}

  {$IFDEF SAK_WSOCK_CANUSESOCKSTREAM}
     while ( tempsockstream.WaitForData(IMAP_MAXSOCKSTREAMWAIT) = False ) do
     begin
        currentTickCount := GetTickCount();
        if ( FSocket.Socket.Connected = False ) then
           raise Exception.create( 'Receive Failed.  No Longer Connected.' );
        if ( currentTickCount - startTickCount > FOperationTimeout ) then
        begin
           raise Exception.create( 'Receive Timeout.  No Data Available.' );
        end;
        Application.ProcessMessages;
        Sleep(0);
     end;
     x := FSocket.Socket.ReceiveBuf( buff, IMAP_BUFFER_SIZE);
  {$ELSE}
     x := FSocket.Socket.ReceiveBuf( buff, IMAP_BUFFER_SIZE);
  {$ENDIF}

  {$IFDEF SAK_WSOCK_CANUSESOCKSTREAM}
  finally
     if assigned( tempsockstream ) then
        tempsockstream.Free();
     // tempsockstream := nil;
  end;
  {$ENDIF}

  FReplyString := copy( buff, 1, x);
  if pos(imap_tag, FReplyString) > 0 then
  begin
    FIMAPError := True;
    FDoErr;
    exit;
  end;

  if pos( imap_tag + ' NO', UpperCase( FReplyString)) = 0 then
  begin
    FSendTextToSocket( RawMail.Text + crlf);
    FReceiveTextFromSocket;
    if pos( imap_tag + ' OK', UpperCase( FReplyString)) > 0 then
    begin
      result := True;
    end;
  end;
end;


function TSakIMAP.RetrieveMessageExt( const MsgIndex : word;
                                      const OnlyHeaders: boolean;
                                      var SakMsg: TSakMsg): boolean;
const
   Nattrstart : string = '(';
   Nattrend   : string = ')';
   Sizestart  : string = '{';
   Sizeend    : string = '}';
   RetFlagsstr : string = 'FLAGS ';
   RetSizestr : string = 'RFC822.SIZE ';
   RetUidstr : string = 'UID ';
   RetRfc822Headerstr : string = 'RFC822.HEADER';
   RetRfc822str : string = 'RFC822 ';
   RetBodypeekstr : string = 'BODY.PEEK[';
var
  msgSize: integer;
  fetchresult : boolean;
  elements : TIMapFetchList;
  tempResultStr : string;
  ponastart, ponaend, currlength : integer;
  currstring : string;
begin
  result := False;
  if not FConnected then
    raise Exception.create( sak_esNotConnected);

  FDoBeforeRetrieve( MsgIndex);

  // === 2001/12/19 - Use fetch method.
  elements := [imapFetchRfc822size];
  If GetIMAPFlagsOnRetrieve then
     Include( elements, imapFetchFlags );
  If GetUIDLsOnRetrieve then
     Include( elements, imapFetchUid );

  fetchresult := Fetch( IntToStr( MsgIndex ), elements );
  if ( fetchresult = False ) then
  begin
     exit;
  end;

  tempResultStr := UpperCase( FReplyString );
  currlength := Length( tempResultStr );

  // Fetch message size
  ponastart := Pos( RetSizestr, tempResultStr );
  ponaend := ponastart;
  while ( ponaend < currlength ) do
  begin
     if ( ( tempResultStr[ponaend+1] = #13 ) or ( tempResultStr[ponaend+1] = #10 ) ) then
        break;
     ponaend := ponaend + 1;
  end;
  currstring := copy(tempResultStr, ponastart, (ponaend - ponastart) + 1);
  msgSize := sak_StrWord2Int( currstring, 2, 100);

  // Fetch IMAP Flags
  if GetIMAPFlagsOnRetrieve then
  begin
    ponastart := Pos( RetFlagsstr, tempResultStr );
    ponaend := ponastart;
    while ( ponaend < currlength ) do
    begin
       if ( ( tempResultStr[ponaend+1] = #13 ) or ( tempResultStr[ponaend+1] = #10 ) ) then
          break;
       ponaend := ponaend + 1;
    end;
    currstring := copy(tempResultStr, ponastart, (ponaend - ponastart) + 1);

    SakMsg.IMAPFlags := [];
    If Pos('\SEEN', currstring)  > 0 then
      SakMsg.IMAPFlags := SakMsg.IMAPFlags + [isSeen];
    If Pos('\ANSWERED', currstring) > 0 then
      SakMsg.IMAPFlags := SakMsg.IMAPFlags + [isAnswered];
    If Pos('\FLAGGED', currstring) > 0 then
      SakMsg.IMAPFlags := SakMsg.IMAPFlags + [isFlagged];
    If Pos('\DELETED', currstring) > 0 then
      SakMsg.IMAPFlags := SakMsg.IMAPFlags + [isDeleted];
    If Pos('\DRAFT', currstring) > 0 then
      SakMsg.IMAPFlags := SakMsg.IMAPFlags + [isDraft];
    If Pos('\RECENT', currstring) > 0 then
      SakMsg.IMAPFlags := SakMsg.IMAPFlags + [isRecent];
  end;    // End Flags

  // Fetch UID
  If GetUIDLsOnRetrieve then
  begin
    ponastart := Pos( RetUidstr, tempResultStr );
    ponaend := ponastart;
    while ( ponaend < currlength ) do
    begin
       if ( ( tempResultStr[ponaend+1] = #13 ) or ( tempResultStr[ponaend+1] = #10 ) ) then
          break;
       ponaend := ponaend + 1;
    end;
    currstring := copy(tempResultStr, ponastart, (ponaend - ponastart) + 1);
    SakMsg.UIDL := IntToStr(sak_StrWord2Int( currstring, 2, 0));
  end;    // End UID

  // ==== 2001/03/21 - Only peek at the messages.  Do not mark as read if
  // ====        automark read is turned off.  My apologies for missing
  // ====        this in the last patch, making AutoMarkRead kind of useless - James.
  // === 2001/12/19 - Use LowLevelFetch method.
  elements := [];
  if OnlyHeaders then
     Include( elements, imapFetchRfc822header )
  else if ( FAutoMarkRead ) then
     Include( elements, imapFetchRfc822 )
  else
     Include( elements, imapFetchBodypeek );

  fetchresult := LowLevelFetch( IntToStr( MsgIndex ), elements, True, False );
  if ( fetchresult = False ) then
  begin
     exit;
  end;

  FReceiveMailFromSocket( SakMsg.RawMail);

  SakMsg.SizeInBytes := msgSize;
  SakMsg.ParseMsg;

  if pos( imap_tag + ' OK', UpperCase( FReplyString)) > 0 then
  begin
    result := True;
  end;

  FDoAfterRetrieve;
end;

function TSakIMAP.RetrieveMessage( const MsgIndex: word;
                                   var SakMsg: TSakMsg): boolean;
begin
  result := RetrieveMessageExt( MsgIndex, False, SakMsg);
end;

function TSakIMAP.RetrieveMessageByUID( const UID: string;
                                        var SakMsg: TSakMsg): boolean;
var
  Idx : integer;
begin
  Idx := FetchMsgIndex(UID);
  if Idx <> 0 then
    result := RetrieveMessageExt( Idx, False, SakMsg)
  else
    result := False;
end;

function TSakIMAP.RetrieveMessageOnlyHeaders( const MsgIndex: word;
                                              var SakMsg: TSakMsg): boolean;
begin
  result := RetrieveMessageExt( MsgIndex, True, SakMsg);
end;

// I have added this function because of the way people leave read messages
// on the IMAP server. Message lists can grow very fast and I often need to
// pull down messages in groups to speed up response times on large mailboxes.
function TSakIMAP.RetrieveRetrieveMessageRangeExt( Start, Finish: integer;
                                               var SakMsgList: TSakMsgList;
                                               const OnlyHeaders: boolean): integer;
var
  i, count: integer;
  aSakMsg: TSakMsg;
begin
  SakMsgList.clear;
  count := 0;
  If Finish > FMsgsCount then
    Finish := FMsgsCount;

  If (Start > Finish) or (Start < 1) then
    begin
    result := -1;
    Exit;
    end;

  for i := Start to Finish do
  begin
     Application.ProcessMessages;
     if FCanceled then
       break;

     aSakMsg := TSakMsg.Create( self);
     if OnlyHeaders then
       RetrieveMessageOnlyHeaders( i, aSakMsg)
     else
       RetrieveMessage( i, aSakMsg);

     SakMsgList.add( aSakMsg);
     inc( count);
  end;
  result := count;
end;

function TSakIMAP.RetrieveMessageRange( Start, Finish: integer;  var SakMsgList: TSakMsgList): integer;
begin
  result := RetrieveRetrieveMessageRangeExt( Start, Finish, SakMsgList, False);
end;

function TSakIMAP.RetrieveMessageRangeOnlyHeaders( Start, Finish: integer; var SakMsgList: TSakMsgList): integer;
begin
  result := RetrieveRetrieveMessageRangeExt( Start, Finish, SakMsgList, True);
end;

function TSakIMAP.RetrieveAllMessagesExt( var SakMsgList: TSakMsgList;
                                          const OnlyHeaders: boolean): integer;
const
   ProcSleepModValue : LongInt = 5;  // === 2001/12/10 - Reduce ProcessMessage requests add a sleep.
var
  i, count: integer;
  aSakMsg: TSakMsg;
  loopCount : LongInt;          // === 2001/12/10 - Track Number of loop elements.
begin
  SakMsgList.clear;
  count := 0;
  loopCount := 0;
  for i := 1 to FMsgsCount do
  begin
     loopCount := loopCount + 1;
     if ( (loopCount mod ProcSleepModValue) = 0 ) then
     begin
        Application.ProcessMessages;
        Sleep(0);
     end;
     if FCanceled then
       break;

     aSakMsg := TSakMsg.Create( self);
     if OnlyHeaders then
       RetrieveMessageOnlyHeaders( i, aSakMsg)
     else
       RetrieveMessage( i, aSakMsg);

     SakMsgList.add( aSakMsg);
     inc( count);
  end;
  result := count;
end;

function TSakIMAP.RetrieveAllMessages( var SakMsgList: TSakMsgList): integer;
begin
  result := RetrieveAllMessagesExt( SakMsgList, False);
end;

function TSakIMAP.RetrieveAllMessagesOnlyHeaders( var SakMsgList: TSakMsgList): integer;
begin
  result := RetrieveAllMessagesExt( SakMsgList, True);
end;

function TSakIMAP.StoreFlags( const MsgIndex: word; Flags: TIMAPFlags): boolean;
// The STORE command alters the Flags associated with the message.
// The \Recent Flag is a read only sstem flag and hence cannot be changed.
var
  cmd, sFlags: string;
begin
  result := False;
  if not FConnected then
    raise Exception.create( sak_esNotConnected);

  imap_tag := ImapGetNextTag(imap_tag);
  If isSeen in Flags then
    sFlags := '\Seen ';
  If isAnswered in Flags then
    sFlags := sFlags + '\Answered ';
  If isFlagged in Flags then
    sFlags := sFlags + '\Flagged ';
  If isDeleted in Flags then
    sFlags := sFlags + '\Deleted ';
  If isDraft in Flags then
    sFlags := sFlags + '\Draft ';

  // The FLAGS argument is used to replace ALL current flags.
  // The SILENT extension is to reduce the response to an OK only.
  cmd := imap_tag + ' STORE ' + IntToStr( MsgIndex) + ' FLAGS.SILENT (' + sFlags + ')' + crlf;

  Application.ProcessMessages;
  if FCanceled then exit;

  FSendTextToSocket( cmd);
  FReceiveTextFromSocket;
  if FIMAPError then
  begin
    FDoErr;
    exit;
  end;

  result := pos( imap_tag + ' OK', UpperCase( FReplyString)) > 0;

  if result then    // === 2001/03/22 - Listen event added - James.
  begin
    FDoStoreFlags( MsgIndex, Flags );
  end;

end;

function TSakIMAP.FetchMsgIndex( const UID: string): word;
var
  cmd : string;
begin
  result := 0;
  if not FConnected then
    raise Exception.create( sak_esNotConnected);

  imap_tag := ImapGetNextTag(imap_tag);
  cmd := imap_tag + ' UID FETCH ' + UID + crlf;

  Application.ProcessMessages;
  if FCanceled then exit;

  FSendTextToSocket( cmd);
  FReceiveTextFromSocket;
  if FIMAPError then
  begin
    FDoErr;
    exit;
  end;
  if pos( imap_tag + ' OK', UpperCase( FReplyString)) > 0 then
    result := sak_StrWord2Int( FReplyString, 2, 0);

end;

function TSakIMAP.SearchRetrieveMessages(Criteria: string; var SakMsgList: TSakMsgList): integer;
begin
  result := SearchRetrieveMessagesExt(Criteria, SakMsgList, False);
end;

function TSakIMAP.SearchRetrieveMessageHeaders(Criteria: string; var SakMsgList: TSakMsgList): integer;
begin
  result := SearchRetrieveMessagesExt(Criteria, SakMsgList, True);
end;

// The SEARCH command takes a LOT of options. I opted for the easy way out.....
function TSakIMAP.SearchRetrieveMessagesExt(Criteria: string;
                                            var SakMsgList: TSakMsgList;
                                            const OnlyHeaders: boolean): integer;
var
  cmd, tmpStr : string;
  i, count: integer;
  aSakMsg: TSakMsg;
  MsgLst : TStringList;
begin
  // Default is none found
  result := 0;
  count := 0;
  SakMsgList.clear;

  if not FConnected then
    raise Exception.create( sak_esNotConnected);

  imap_tag := ImapGetNextTag(imap_tag);

  cmd := imap_tag + ' SEARCH ' + Criteria + crlf;

  Application.ProcessMessages;
  if FCanceled then exit;

  FSendTextToSocket( cmd);
  FReceiveTextFromSocket;
  if FIMAPError then
  begin
    FDoErr;
    exit;
  end;

  if pos( imap_tag + ' OK', UpperCase( FReplyString)) > 0 then
    if pos( '* SEARCH', UpperCase( FReplyString)) > 0 then
    begin
      MsgLst := TStringList.Create;
      try
        tmpStr := Copy(FReplyString, 10, Pos(#$D#$A, FReplyString)-10);
        tmpStr := StringReplace( tmpStr, ' ', ',', [rfReplaceAll, rfIgnoreCase]);
        MsgLst.CommaText := tmpStr;
        if MsgLst.Count > 0 then
          for i := 0 to MsgLst.Count-1 do
          begin
            Application.ProcessMessages;
            if FCanceled then
              break;

            aSakMsg := TSakMsg.Create( self);
            if OnlyHeaders then
              RetrieveMessageOnlyHeaders( StrToInt(MsgLst[i]), aSakMsg)
            else
              RetrieveMessage( StrToInt(MsgLst[i]), aSakMsg);

            SakMsgList.add( aSakMsg);
            inc( count);
          end;
      finally
        MsgLst.Free;
      end;
    end;
  result := count;
end;


// === 2001/02/07 - New mark read method - James.
// === 2001/03/22 - Replaced by enumeration based set and clear methods - James.
function TSakIMAP.SetMessageFlag( const MsgIndex: word; setFlag : TIMAPFlag ) : boolean;
// you must select a folder first
// and following this method the message at MsgIndex will be have setFlag set
// pending a commit.
var
  cmd: string;
begin
  result := False;
  if not FConnected then
    raise Exception.create( sak_esNotConnected);

  imap_tag := ImapGetNextTag(imap_tag);
  cmd := imap_tag + ' STORE ' + IntToStr( MsgIndex) + ' +FLAGS ';
  case setFlag of
    isSeen: cmd := cmd + '(\seen)';
    isAnswered: cmd := cmd + '(\answered)';
    isFlagged: cmd := cmd + '(\flagged)';
    isDeleted: cmd := cmd + '(\deleted)';
    isDraft: cmd := cmd + '(\draft)';
    isRecent: cmd := cmd + '(\recent)';
  else
     raise Exception.create( sak_esInvalidOp );
  end;

  cmd := cmd + crlf;

  Application.ProcessMessages;
  if FCanceled then exit;

  FSendTextToSocket( cmd);
  FReceiveTextFromSocket;

  if FIMAPError then
  begin
    FDoErr;
    exit;
  end;

  result := pos( imap_tag + ' OK', UpperCase( FReplyString)) > 0;
  if result then
  begin
    FDoSetFlag( MsgIndex, setFlag );
  end;
end;

// === 2001/03/22 - Replaced by enumeration based set and clear methods - James.
function TSakIMAP.ClearMessageFlag( const MsgIndex: word; setFlag : TIMAPFlag ) : boolean;
// you must select a folder first
// and following this method the message at MsgIndex will be have setFlag cleared
// pending a commit.
var
  cmd: string;
begin
  result := False;
  if not FConnected then
    raise Exception.create( sak_esNotConnected);

  imap_tag := ImapGetNextTag(imap_tag);
  cmd := imap_tag + ' STORE ' + IntToStr( MsgIndex) + ' -FLAGS ';
  case setFlag of
    isSeen: cmd := cmd + '(\seen)';
    isAnswered: cmd := cmd + '(\answered)';
    isFlagged: cmd := cmd + '(\flagged)';
    isDeleted: cmd := cmd + '(\deleted)';
    isDraft: cmd := cmd + '(\draft)';
    isRecent: cmd := cmd + '(\recent)';
  else
     raise Exception.create( sak_esInvalidOp );
  end;

  cmd := cmd + crlf;

  Application.ProcessMessages;
  if FCanceled then exit;

  FSendTextToSocket( cmd);
  FReceiveTextFromSocket;

  if FIMAPError then
  begin
    FDoErr;
    exit;
  end;

  result := pos( imap_tag + ' OK', UpperCase( FReplyString)) > 0;
  if result then
  begin
    FDoClearFlag( MsgIndex, setFlag );
  end;
end;


// === 2001/02/08 - New Checkpoint method - James.
function TSakIMAP.DoCheckpoint(): boolean;
// you must select a folder first
// and following this method the mailbox checkpoint should occur for the active folder.
var
  cmd: string;
begin
  result := False;
  if not FConnected then
    raise Exception.create( sak_esNotConnected);

  imap_tag := ImapGetNextTag(imap_tag);
  cmd := imap_tag + ' CHECK' + crlf;

  Application.ProcessMessages;
  if FCanceled then exit;

  FSendTextToSocket( cmd);
  FReceiveTextFromSocket;

  if FIMAPError then
  begin
    FDoErr;
    exit;
  end;

  result := pos( imap_tag + ' OK', UpperCase( FReplyString)) > 0;
  if result then
  begin
    FDoCheckpoint();
  end;
end;

// === 2001/02/08 - New SetACL method - James.
// === 2001/03/22 - Updated to utilize an enumerated type - James.
function TSakIMAP.SetACL( const folder: string; const identifier: string; rights : TIMAPACLRights ) : boolean;
// you must select a folder first
// and following this method the ACL changes should occur for the current mailbox and identifier.
var
  cmd: string;
  tmpFolder : string; // === 2001/12/10
begin
  result := False;
  if not FACLExtensionsOn then
    raise Exception.create( 'ACL Extensions Not Active' );

  if not FConnected then
    raise Exception.create( sak_esNotConnected);

  imap_tag := ImapGetNextTag(imap_tag);
  tmpFolder := FixFolderName( folder );  // === 2001/12/10
  cmd := imap_tag + ' SETACL ' + tmpFolder + ' ' + identifier + ' +';  // === 2001/12/10

  If imapLookup in rights then
    cmd := cmd + TIMAPACLRightsChar[ord(imapLookup)];
  If imapRead in rights then
    cmd := cmd + TIMAPACLRightsChar[ord(imapRead)];
  If imapSeen in rights then
    cmd := cmd + TIMAPACLRightsChar[ord(imapSeen)];
  If imapWrite in rights then
    cmd := cmd + TIMAPACLRightsChar[ord(imapWrite)];
  If imapInsert in rights then
    cmd := cmd + TIMAPACLRightsChar[ord(imapInsert)];
  If imapPost in rights then
    cmd := cmd + TIMAPACLRightsChar[ord(imapPost)];
  If imapCreate in rights then
    cmd := cmd + TIMAPACLRightsChar[ord(imapCreate)];
  If imapDelete in rights then
    cmd := cmd + TIMAPACLRightsChar[ord(imapDelete)];
  If imapAdminister in rights then
    cmd := cmd + TIMAPACLRightsChar[ord(imapAdminister)];

  cmd := cmd + crlf;

  Application.ProcessMessages;
  if FCanceled then exit;

  FSendTextToSocket( cmd);
  FReceiveTextFromSocket;

  if FIMAPError then
  begin
    FDoErr;
    exit;
  end;

  result := pos( imap_tag + ' OK', UpperCase( FReplyString)) > 0;
  if result then
  begin
    FDoSetACL(folder, identifier, rights);
  end;
end;

// === 2001/03/22 - Updated to utilize an enumerated type - James.
function TSakIMAP.ClearACL( const folder: string; const identifier: string; rights : TIMAPACLRights ) : boolean;
// you must select a folder first
// and following this method the ACL changes should occur for the current mailbox and identifier.
var
  cmd: string;
  tmpFolder : string; // === 2001/12/10
begin
  result := False;
  if not FACLExtensionsOn then
    raise Exception.create( 'ACL Extensions Not Active' );

  if not FConnected then
    raise Exception.create( sak_esNotConnected);

  imap_tag := ImapGetNextTag(imap_tag);
  tmpFolder := FixFolderName( folder );  // === 2001/12/10
  cmd := imap_tag + ' SETACL ' + tmpFolder + ' ' + identifier + ' -';  // == 2001/12/10

  If imapLookup in rights then
    cmd := cmd + TIMAPACLRightsChar[ord(imapLookup)];
  If imapRead in rights then
    cmd := cmd + TIMAPACLRightsChar[ord(imapRead)];
  If imapSeen in rights then
    cmd := cmd + TIMAPACLRightsChar[ord(imapSeen)];
  If imapWrite in rights then
    cmd := cmd + TIMAPACLRightsChar[ord(imapWrite)];
  If imapInsert in rights then
    cmd := cmd + TIMAPACLRightsChar[ord(imapInsert)];
  If imapPost in rights then
    cmd := cmd + TIMAPACLRightsChar[ord(imapPost)];
  If imapCreate in rights then
    cmd := cmd + TIMAPACLRightsChar[ord(imapCreate)];
  If imapDelete in rights then
    cmd := cmd + TIMAPACLRightsChar[ord(imapDelete)];
  If imapAdminister in rights then
    cmd := cmd + TIMAPACLRightsChar[ord(imapAdminister)];

  cmd := cmd + crlf;

  Application.ProcessMessages;
  if FCanceled then exit;

  FSendTextToSocket( cmd);
  FReceiveTextFromSocket;

  if FIMAPError then
  begin
    FDoErr;
    exit;
  end;

  result := pos( imap_tag + ' OK', UpperCase( FReplyString)) > 0;
  if result then
  begin
    FDoClearACL(folder, identifier, rights);
  end;
end;

// === 2001/02/08 - New DeleteACL method - James.
function TSakIMAP.DeleteACL( const folder: string; const identifier: string ) : boolean;
// you must select a folder first
// and following this method the ACL changes should occur for the current mailbox and identifier.
var
  cmd: string;
  tmpFolder : string; // === 2001/12/10
begin
  result := False;
  if not FACLExtensionsOn then
    raise Exception.create( 'ACL Extensions Not Active' );

  if not FConnected then
    raise Exception.create( sak_esNotConnected);

  imap_tag := ImapGetNextTag(imap_tag);
  tmpFolder := FixFolderName( folder );  // === 2001/12/10
  cmd := imap_tag + ' DELETEACL ' + tmpFolder + ' ' + identifier + crlf;  // == 2001/12/10

  Application.ProcessMessages;
  if FCanceled then exit;

  FSendTextToSocket( cmd);
  FReceiveTextFromSocket;

  if FIMAPError then
  begin
    FDoErr;
    exit;
  end;

  result := pos( imap_tag + ' OK', UpperCase( FReplyString)) > 0;
  if result then
  begin
    FDoDeleteACL( folder, identifier );
  end;
end;

// === 2001/02/08 - New GetACL method - James.
function TSakIMAP.GetACL( const folder : string ) : boolean;
// you must select a folder first
// and following this method the ACL information is returned.
var
  cmd: string;
  tmpFolder : string; // === 2001/12/10
begin
  result := False;
  if not FACLExtensionsOn then
    raise Exception.create( 'ACL Extensions Not Active' );

  if not FConnected then
    raise Exception.create( sak_esNotConnected);

  imap_tag := ImapGetNextTag(imap_tag);
  tmpFolder := FixFolderName( folder );  // === 2001/12/10
  cmd := imap_tag + ' GETACL ' + tmpFolder + crlf;   // === 2001/12/10

  Application.ProcessMessages;
  if FCanceled then exit;

  FSendTextToSocket( cmd);
  FReceiveTextFromSocket;

  if FIMAPError then
  begin
    FDoErr;
    exit;
  end;

  result := pos( imap_tag + ' OK', UpperCase( FReplyString)) > 0;
  if result then
  begin
    FDoGetACL( folder );
  end;
end;

// === 2001/03/22 - New ListrightsACL method - James.
function TSakIMAP.ListrightsACL( const folder: string; const identifier: string ) : boolean;
// you must select a folder first
// and following this method the ACL changes should occur for the current mailbox and identifier.
var
  cmd: string;
  tmpFolder : string; // === 2001/12/10
begin
  result := False;
  if not FACLExtensionsOn then
    raise Exception.create( 'ACL Extensions Not Active' );

  if not FConnected then
    raise Exception.create( sak_esNotConnected);

  imap_tag := ImapGetNextTag(imap_tag);
  tmpFolder := FixFolderName( folder );  // === 2001/12/10
  cmd := imap_tag + ' LISTRIGHTS ' + tmpFolder + ' ' + identifier + crlf;  // === 2001/12/10

  Application.ProcessMessages;
  if FCanceled then exit;

  FSendTextToSocket( cmd);
  FReceiveTextFromSocket;

  if FIMAPError then
  begin
    FDoErr;
    exit;
  end;

  result := pos( imap_tag + ' OK', UpperCase( FReplyString)) > 0;
  if result then
  begin
    FDoListrightsACL( folder, identifier );
  end;
end;

// === 2001/03/22 - New MyrightsACL method - James.
function TSakIMAP.MyrightsACL( const folder: string; const identifier: string ) : boolean;
// you must select a folder first
// and following this method the ACL information is returned.
var
  cmd: string;
  tmpFolder : string; // === 2001/12/10
begin
  result := False;
  if not FACLExtensionsOn then
    raise Exception.create( 'ACL Extensions Not Active' );

  if not FConnected then
    raise Exception.create( sak_esNotConnected);

  imap_tag := ImapGetNextTag(imap_tag);
  tmpFolder := FixFolderName( folder );  // === 2001/12/10
  cmd := imap_tag + ' MYRIGHTS ' + tmpfolder + crlf;

  Application.ProcessMessages;
  if FCanceled then exit;

  FSendTextToSocket( cmd);
  FReceiveTextFromSocket;

  if FIMAPError then
  begin
    FDoErr;
    exit;
  end;

  result := pos( imap_tag + ' OK', UpperCase( FReplyString)) > 0;
  if result then
  begin
    FDoMyrightsACL( folder );
  end;
end;

// === 2001/02/09 - Rename Folder Method added - James.
function TSakIMAP.RenameFolder(const OriginalFolderName: string; const NewFolderName : string ): boolean;
var
  cmd: string;
  tmpFolder, tmpFolder2 : string; // === 2001/12/10
begin
  //result := False;
  imap_tag := ImapGetNextTag(imap_tag);
  tmpFolder := FixFolderName( OriginalFolderName );  // === 2001/12/10
  tmpFolder2 := FixFolderName( NewFolderName );      // === 2001/12/10

  cmd := imap_tag + ' RENAME ' + tmpFolder + ' ' + tmpFolder2 + crlf;   // === 2001/12/10

  FSendTextToSocket( cmd);
  FReceiveTextFromSocket;

  if FIMAPError then
  begin
    result := False;
    FDoErr;
    exit;
  end;

  result := pos( imap_tag + ' OK', UpperCase( FReplyString)) > 0;
  // you can inspect ReplyString to see what happened in case
  // of returning False
end;


// === 2001/02/09 - Subscribe Folder Method added - James.
function TSakIMAP.SubscribeFolder(const FolderName : string ) : boolean;
var
  cmd: string;
  tmpFolder : string; // === 2001/12/10
begin
  //result := False;
  imap_tag := ImapGetNextTag(imap_tag);
  tmpFolder := FixFolderName( FolderName );  // === 2001/12/10
  cmd := imap_tag + ' SUBSCRIBE ' + tmpFolder +  crlf;  // === 2001/12/10

  FSendTextToSocket( cmd);
  FReceiveTextFromSocket;

  if FIMAPError then
  begin
    result := False;
    FDoErr;
    exit;
  end;

  result := pos( imap_tag + ' OK', UpperCase( FReplyString)) > 0;
  // you can inspect ReplyString to see what happened in case
  // of returning False
end;

// === 2001/02/09 - Unsubscribe Folder Method added - James.
function TSakIMAP.UnsubscribeFolder(const FolderName : string ) : boolean;
var
  cmd: string;
  tmpFolder : string; // === 2001/12/10
begin
  //result := False;
  imap_tag := ImapGetNextTag(imap_tag);
  tmpFolder := FixFolderName( FolderName );  // === 2001/12/10
  cmd := imap_tag + ' UNSUBSCRIBE ' + tmpFolder +  crlf;  // === 2001/12/10

  FSendTextToSocket( cmd);
  FReceiveTextFromSocket;

  if FIMAPError then
  begin
    result := False;
    FDoErr;
    exit;
  end;

  result := pos( imap_tag + ' OK', UpperCase( FReplyString)) > 0;
  // you can inspect ReplyString to see what happened in case
  // of returning False
end;

// === 2001/02/09 - Subscribe Folder Method added - James.
function TSakIMAP.LSUBFolder(const ReferenceName : string; const WildcardMailbox : string ) : boolean;
var
  cmd: string;
begin
  //result := False;
  imap_tag := ImapGetNextTag(imap_tag);
  cmd := imap_tag + ' LSUB ' + ReferenceName + ' ' + WildcardMailbox +  crlf;

  FSendTextToSocket( cmd);
  FReceiveTextFromSocket;

  if FIMAPError then
  begin
    result := False;
    FDoErr;
    exit;
  end;

  result := pos( imap_tag + ' OK', UpperCase( FReplyString)) > 0;
  // you can inspect ReplyString to see what happened in case
  // of returning False
end;

// === 2001/02/09 - Copy message(s) from the current folder to the destination - James.
function TSakIMAP.CopyMessageToFolder(const MessageSet : string; const FolderName : string ) : boolean;
begin
  result := LowLevelCopy( MessageSet, FolderName, False );  // ==== 2001/12/21 - Use lowlevel
end;

// ==== 2001/12/18 - Attempt to round out IMAP4rev1 command capabilities ===
function TSakIMAP.Capability() : boolean;
var
  cmd: string;
begin
  //result := False;
  imap_tag := ImapGetNextTag(imap_tag);
  cmd := imap_tag + ' CAPABILITY' + crlf;

  FSendTextToSocket( cmd );
  FReceiveTextFromSocket;

  if FIMAPError then
  begin
    result := False;
    FDoErr;
    exit;
  end;

  result := pos( imap_tag + ' OK', UpperCase( FReplyString)) > 0;
  if result then
  begin
    FDoCapability();
  end;
end;

// ==== 2001/12/18 - Attempt to round out IMAP4rev1 command capabilities ===
function TSakIMAP.Noop() : boolean;
const
   LineStart : Char = '*';
   NoMessagesstr : string = ' EXISTS';
   NoRecentstr : string = ' RECENT';
var
  cmd: string;
  ponastart, ponaend : integer;
  nattrstring, currstring : string;
begin
  //result := False;
  imap_tag := ImapGetNextTag(imap_tag);
  cmd := imap_tag + ' NOOP' + crlf;

  FSendTextToSocket( cmd );
  FReceiveTextFromSocket;

  if FIMAPError then
  begin
    result := False;
    FDoErr;
    exit;
  end;

  result := pos( imap_tag + ' OK', UpperCase( FReplyString)) > 0;
  if result then
  begin
    nattrstring := UpperCase( FReplyString );
    ponaend := Pos( NoRecentstr, nattrstring );
    if ( ponaend > 0 ) then
    begin
       ponastart := ponaend;
       while ( ponastart > 0 ) do
       begin
          if nattrstring[ponastart] = LineStart then
             break;
          ponastart := ponastart - 1;
       end;
       if ponastart <= 0 then
          ponastart := 1;
       currstring := copy(nattrstring, ponastart, (ponaend - ponastart) + 1);
       FRecentMsgsCount := sak_StrWord2Int( currstring, 2, 0);
    end;
    ponaend := Pos( NoMessagesstr, nattrstring );
    if ( ponaend > 0 ) then
    begin
       ponastart := ponaend;
       while ( ponastart > 0 ) do
       begin
          if nattrstring[ponastart] = LineStart then
             break;
          ponastart := ponastart - 1;
       end;
       if ponastart <= 0 then
          ponastart := 1;
       currstring := copy(nattrstring, ponastart, (ponaend - ponastart) + 1);
       FMsgsCount := sak_StrWord2Int( currstring, 2, 0);
    end;

    FDoNoop(FSelectedFolder, FMsgsCount, FRecentMsgsCount);
  end;
end;

// === 2001/12/19 - Status Request - any folder ===
function TSakIMAP.Status( const folder: string; elements: TIMAPStatusList ) : boolean;
const
   Nattrstart : string = '(';
   Nattrend   : string = ')';
var
  cmd: string;
  tmpFolder : string;
begin
  result := False;
  if not FConnected then
    raise Exception.create( sak_esNotConnected);

  imap_tag := ImapGetNextTag(imap_tag);
  tmpFolder := FixFolderName( folder );
  cmd := imap_tag + ' STATUS ' + tmpFolder + ' ' + Nattrstart;

  If imapMessages in elements then
    cmd := cmd + TIMAPStatusListStrings[ord(imapMessages)] + ' ';
  If imapRecent in elements then
    cmd := cmd + TIMAPStatusListStrings[ord(imapRecent)] + ' ';
  If imapUidnext in elements then
    cmd := cmd + TIMAPStatusListStrings[ord(imapUidnext)] + ' ';
  If imapUidvalidity in elements then
    cmd := cmd + TIMAPStatusListStrings[ord(imapUidvalidity)] + ' ';
  If imapUnseen in elements then
    cmd := cmd + TIMAPStatusListStrings[ord(imapUnseen)] + ' ';
  cmd := TrimRight( cmd );
  cmd := cmd + Nattrend + crlf;

  Application.ProcessMessages;
  if FCanceled then exit;

  FSendTextToSocket( cmd);
  FReceiveTextFromSocket;

  if FIMAPError then
  begin
    FDoErr;
    exit;
  end;

  result := pos( imap_tag + ' OK', UpperCase( FReplyString)) > 0;
  if result then
  begin
    FDoStatus(folder, elements);
  end;
end;

// === 2001/12/19 - Fetch Request - selected folder ===
function TSakIMAP.LowLevelFetch( const messageset: string; elements: TIMAPFetchList;
                                 const sendonly: boolean; const UIDfetch: boolean ) : boolean;
const
   Nattrstart : string = '(';
   Nattrend   : string = ')';
var
  cmd: string;
begin
  result := False;
  if not FConnected then
    raise Exception.create( sak_esNotConnected);

  imap_tag := ImapGetNextTag(imap_tag);

  if ( UIDFetch = True ) then
     cmd := imap_tag + ' UID FETCH ' + messageset + ' ' + Nattrstart
  else
     cmd := imap_tag + ' FETCH ' + messageset + ' ' + Nattrstart;

  If imapFetchFast in elements then
    cmd := cmd + TIMAPFetchListStrings[ord(imapFetchFast)] + ' ';
  If imapFetchAll in elements then
    cmd := cmd + TIMAPFetchListStrings[ord(imapFetchAll)] + ' ';
  If imapFetchFull in elements then
    cmd := cmd + TIMAPFetchListStrings[ord(imapFetchFull)] + ' ';
  If imapFetchUid in elements then
    cmd := cmd + TIMAPFetchListStrings[ord(imapFetchUid)] + ' ';
  If imapFetchFlags in elements then
    cmd := cmd + TIMAPFetchListStrings[ord(imapFetchFlags)] + ' ';
  If imapFetchInternalDate in elements then
    cmd := cmd + TIMAPFetchListStrings[ord(imapFetchInternalDate)] + ' ';
  If imapFetchEnvelope in elements then
    cmd := cmd + TIMAPFetchListStrings[ord(imapFetchEnvelope)] + ' ';
  If imapFetchBodystructure in elements then
    cmd := cmd + TIMAPFetchListStrings[ord(imapFetchBodystructure)] + ' ';
  If imapFetchBody in elements then
    cmd := cmd + TIMAPFetchListStrings[ord(imapFetchBody)] + ' ';
  If imapFetchBodysection in elements then
    cmd := cmd + TIMAPFetchListStrings[ord(imapFetchBodysection)] + '[]' + ' ';
  If imapFetchBodypeek in elements then
    cmd := cmd + TIMAPFetchListStrings[ord(imapFetchBodypeek)] + '[]' + ' ';
  If imapFetchRfc822 in elements then
    cmd := cmd + TIMAPFetchListStrings[ord(imapFetchRfc822)] + ' ';
  If imapFetchRfc822header in elements then
    cmd := cmd + TIMAPFetchListStrings[ord(imapFetchRfc822header)] + ' ';
  If imapFetchRfc822size in elements then
    cmd := cmd + TIMAPFetchListStrings[ord(imapFetchRfc822size)] + ' ';
  If imapFetchRfc822text in elements then
    cmd := cmd + TIMAPFetchListStrings[ord(imapFetchRfc822text)] + ' ';
  cmd := TrimRight( cmd );
  cmd := cmd + Nattrend + crlf;

  Application.ProcessMessages;
  if FCanceled then exit;

  FSendTextToSocket( cmd);
  if ( sendonly = False ) then
  begin
     FReceiveTextFromSocket;

     if FIMAPError then
     begin
       FDoErr;
       exit;
     end;

     result := pos( imap_tag + ' OK', UpperCase( FReplyString)) > 0;
     if result then
     begin
       if ( UIDFetch = True ) then
          FDoFetchUID(FSelectedFolder, messageset, elements)
       else
          FDoFetch(FSelectedFolder, messageset, elements);
     end;
  end
  else
  begin
     result := True;
  end;
end;

// === 2001/12/19 - Fetch Request - selected folder ===
function TSakIMAP.Fetch( const messageset: string; elements: TIMAPFetchList ) : boolean;
begin
  result := LowLevelFetch( messageset, elements, False, False );
end;


// === 2001/12/19 - Fetch Body Request - selected folder ===
function TSakIMAP.FetchBody( const messageset: string; const peekonly: boolean;
                             const sectionstring: string; const partialstring: string ) : boolean;
const
   Nattrstart : string = '(';
   Nattrend   : string = ')';
   Sattrstart : string = '[';
   Sattrend   : string = ']';
   Pattrstart : string = '<';
   Pattrend   : string = '>';
var
  cmd: string;
  elements : TIMAPFetchList;
begin
  result := False;
  if not FConnected then
    raise Exception.create( sak_esNotConnected);

  if ( peekonly = True ) then
     elements := [imapFetchBodypeek]
  else if ( sectionstring = '') then
     elements := [imapFetchBody]
  else
     elements := [imapFetchBodysection];

  imap_tag := ImapGetNextTag(imap_tag);
  cmd := imap_tag + ' FETCH ' + messageset + ' ' + Nattrstart;

  if imapFetchBody in elements then
  begin
    cmd := cmd + TIMAPFetchListStrings[ord(imapFetchBody)];
  end;
  if imapFetchBodysection in elements then
  begin
    cmd := cmd + TIMAPFetchListStrings[ord(imapFetchBodysection)] + Sattrstart + sectionstring +
                 Sattrend;
    if ( partialstring <> '' ) then
       cmd := cmd + Pattrstart + partialstring + Pattrend;
  end;
  if imapFetchBodypeek in elements then
  begin
    cmd := cmd + TIMAPFetchListStrings[ord(imapFetchBodypeek)] + Sattrstart;
    if ( sectionstring <> '' ) then
    begin
       cmd := cmd + sectionstring + Sattrend;
       if ( partialstring <> '' ) then
          cmd := cmd + Pattrstart + partialstring + Pattrend;
    end
    else
       cmd := cmd + Sattrend;
  end;
  cmd := TrimRight( cmd );
  cmd := cmd + Nattrend + crlf;

  Application.ProcessMessages;
  if FCanceled then exit;

  FSendTextToSocket( cmd);
  FReceiveTextFromSocket;

  if FIMAPError then
  begin
    FDoErr;
    exit;
  end;

  result := pos( imap_tag + ' OK', UpperCase( FReplyString)) > 0;
  if result then
  begin
    FDoFetchBody(FSelectedFolder, messageset, elements, sectionstring, partialstring);
  end;
end;

// === 2001/12/19 - Examine Request - select folder in readonly mode ===
function TSakIMAP.ExamineFolder(const FolderName: string): boolean;
const
   Nattrstart : string = '(';
   Nattrend   : string = ')';
   NoMessagesstr : string = 'MESSAGES ';
   NoRecentstr : string = 'RECENT ';
   NoUnseenstr : string = 'UNSEEN ';
var
  cmd, tmpFolder: string;
  ponastart, ponaend, currpos : integer;
  nattrstring, currstring : string;
  statusresult : boolean;
begin
  //result := False;
  tmpFolder := trim(FolderName);
  tmpFolder := FixFolderName( tmpFolder );  // === 2001/12/10

  // === 2002/01/03 - See SelectFolder comments for anation of event ordering - James.
  if ( FSelectedFolder <> '' ) then
  begin
     if ( CloseSelectedFolder() = False ) then
     begin
        // === 2002/01/03 - A state problem exists - client should disconnect to clear.
        //result := False;
        raise Exception.create( 'Examine Failed.  Cannot close previously selected folder.' +
                                #13#10 + 'Disconnect from server if error persists.' );
        exit;
     end;
  end;

  // === 2002/01/03 - Before Examine, update MsgCount, NewMsgsCount and RecentMsgsCount.
  // === 2001/12/19 - Results set to 0 on failure.
  FMsgsCount := 0;
  FNewMsgsCount := 0;
  FRecentMsgsCount := 0;

  statusresult := Status(FolderName, [imapMessages, imapUnseen, imapRecent]);

  if ( statusresult = True ) then
  begin
     ponastart := Pos( Nattrstart, FReplyString );
     ponaend := Pos( Nattrend, FReplyString );
     if ( ponastart > 0 ) and ( ponaend > 0 ) and ( ponaend > ponastart ) then
     begin
        nattrstring := copy(FReplyString, ponastart, (ponaend - ponastart) + 1);
        nattrstring := UpperCase(nattrstring);
        currpos := Pos( NoMessagesstr, nattrstring );
        if ( currpos > 0 ) then
        begin
           currstring := copy(nattrstring, currpos, (Length(nattrstring) - currpos) + 1);
           FMsgsCount := sak_StrWord2Int( currstring, 2, 0);
        end;
        currpos := Pos( NoUnseenstr, nattrstring );
        if ( currpos > 0 ) then
        begin
           currstring := copy(nattrstring, currpos, (Length(nattrstring) - currpos) + 1);
           FNewMsgsCount := sak_StrWord2Int( currstring, 2, 0);
        end;
        currpos := Pos( NoRecentstr, nattrstring );
        if ( currpos > 0 ) then
        begin
           currstring := copy(nattrstring, currpos, (Length(nattrstring) - currpos) + 1);
           FRecentMsgsCount := sak_StrWord2Int( currstring, 2, 0);
        end;
     end;
  end;

  imap_tag := ImapGetNextTag(imap_tag);
  cmd := imap_tag + ' EXAMINE ' + tmpFolder + crlf;

  FSendTextToSocket( cmd);
  FReceiveTextFromSocket;

  if FIMAPError then
  begin
    result := False;
    FDoErr;
    exit;
  end;

  result := pos( imap_tag + ' OK', UpperCase( FReplyString)) > 0;

  if not result then
    exit;
  // you can inspect ReplyString to see what happened in case
  // of returning False

  FSelectedFolder := trim(FolderName);   // === 2001/12/10
end;

// === 2001/12/20 - Close selected folder ===
function TSakIMAP.CloseSelectedFolder(): boolean;
var
  cmd: string;
begin
  //result := False;
  imap_tag := ImapGetNextTag(imap_tag);
  cmd := imap_tag + ' CLOSE' + crlf;

  FSendTextToSocket(cmd);
  FReceiveTextFromSocket;

  if FIMAPError then
  begin
    result := False;
    FDoErr;
    exit;
  end;

  result := pos( imap_tag + ' OK', UpperCase( FReplyString)) > 0;
  if ( result = True ) then
  begin
     // == Clear selected folder and counts on successful close. ==
     FSelectedFolder := '';
     FMsgsCount := 0;
     FNewMsgsCount := 0;
     FRecentMsgsCount := 0;
  end;
end;

// ==== 2001/12/18 - Attempt to round out IMAP4rev1 command capabilities ===
function TSakIMAP.Idle( timeout: word ) : boolean;
const
   ProcSleepModValue : LongInt = 5;  // === 2001/12/10 - Reduce ProcessMessage requests add a sleep.
   LineStart : Char = '*';
   NoMessagesstr : string = ' EXISTS';
   NoRecentstr : string = ' RECENT';
   DoneStr : string = 'DONE';
   MaxIDLE : LongInt = 1710000; // === 2001/12/20 - <= 29 minutes ( RFC2177 ) - James
var
  cmd: string;
  ponastart, ponaend, endfound : integer;
  nattrstring, currstring : string;
  s, l, buff_tail: string;      // === 2001/12/20 - Use FReceiveMailFromSocket() lbt processing
  x, lbt, pos_tag: integer;     // === 2001/12/20 - Use FReceiveMailFromSocket() lbt processing
  buff: array[1..IMAP_BUFFER_SIZE] of char;
  startTickCount : LongInt;     // === 2001/12/10 - Track Start of operation
  currentTickCOunt : LongInt;   // === 2001/12/10 - Track Current ticks
  loopCount, loopCount2 : LongInt;          // === 2001/12/10 - Track Number of loop elements.
  doneRequested : boolean;
{$IFDEF SAK_WSOCK_CANUSESOCKSTREAM}
  tempsockstream : TWinSocketStream;
  sockstreamwait : LongInt;
{$ENDIF}
begin
  result := False;
  imap_tag := ImapGetNextTag(imap_tag);
  cmd := imap_tag + ' IDLE' + crlf;
  FIdleDonerequested := False;
  doneRequested := False;

  FSendTextToSocket( cmd );

  s := '';
  buff_tail := '';
  loopCount := 0;
  pos_tag := 0;
  if ( timeout > MaxIDLE ) then
     timeout := MaxIDLE;

  {$IFDEF SAK_WSOCK_CANUSESOCKSTREAM}
  tempsockstream := nil;
  try
     sockstreamwait := ( timeout div 4 );
     if ( sockstreamwait > IMAP_MAXSOCKSTREAMWAIT ) then
        sockstreamwait := IMAP_MAXSOCKSTREAMWAIT;
     if ( sockstreamwait < 100 ) then
        sockstreamwait := 100;
     tempsockstream := TWinSocketStream.Create( FSocket.Socket, sockstreamwait );
  {$ENDIF}

  startTickCount := GetTickCount();

  while True do
  begin
    loopCount := loopCount + 1;
    if ( (loopCount mod ProcSleepModValue) = 0 ) then
    begin
       Application.ProcessMessages;
       Sleep(0);
    end;

    {$IFDEF SAK_WSOCK_CANUSESOCKSTREAM}
       if ( tempsockstream.WaitForData(sockstreamwait) = True ) then
          x := FSocket.Socket.ReceiveBuf( buff, IMAP_BUFFER_SIZE)
       else
       begin
          x := 0; // === No data received.
          if ( FSocket.Socket.Connected = False ) then
             raise Exception.create( 'Receive Failed.  No Longer Connected.' );
       end;
    {$ELSE}
       x := FSocket.Socket.ReceiveBuf( buff, IMAP_BUFFER_SIZE);
    {$ENDIF}

    if ( x <= 0 ) then
    begin
       Application.ProcessMessages;
       Sleep(0);
    end
    else
    begin
       s := s + copy( buff, 1, x);

       // === 2001/12/20 - Use same mail message lbt processing.  Same logic.
       if x > 99 then
         buff_tail := copy( buff, x-99, 100)
       else
         buff_tail := buff_tail + copy( buff, 1, x);

       // === Update Message counts while idling ===
       endfound := 0;
       nattrstring := UpperCase( buff_tail );
       ponaend := Pos( NoRecentstr, nattrstring );
       if ( ponaend > 0 ) then
       begin
          endfound := ponaend;
          ponastart := ponaend;
          while ( ponastart > 0 ) do
          begin
             if nattrstring[ponastart] = LineStart then
                break;
             ponastart := ponastart - 1;
          end;
          if ponastart <= 0 then
             ponastart := 1;
          currstring := copy(nattrstring, ponastart, (ponaend - ponastart) + 1);
          FRecentMsgsCount := sak_StrWord2Int( currstring, 2, 0);
          FDoIdle(FSelectedFolder, FMsgsCount, FRecentMsgsCount, False);
       end;
       ponaend := Pos( NoMessagesstr, nattrstring );
       if ( ponaend > 0 ) then
       begin
          if ( ponaend > endfound ) then
             endfound := ponaend;
          ponastart := ponaend;
          while ( ponastart > 0 ) do
          begin
             if nattrstring[ponastart] = LineStart then
                break;
             ponastart := ponastart - 1;
          end;
          if ponastart <= 0 then
             ponastart := 1;
          currstring := copy(nattrstring, ponastart, (ponaend - ponastart) + 1);
          FMsgsCount := sak_StrWord2Int( currstring, 2, 0);
          FDoIdle(FSelectedFolder, FMsgsCount, FRecentMsgsCount, False);
       end;

       pos_tag := pos( imap_tag, buff_tail);
       // if we found "imap_tag * crlf" then we have finished
       if pos_tag > 0 then
       begin
         l := copy( buff_tail, pos_tag, length(buff_tail));
         if pos(crlf, l) > 0 then
           break;
       end;

       lbt := length(buff_tail);
       if ( endfound > 0 ) then
          buff_tail := copy( buff_tail, endfound + 1, lbt );
       if lbt > 99 then
         buff_tail := copy(buff_tail, lbt-40, lbt);
    end; // if x<=0

    currentTickCount := GetTickCount();
    if ( ( currentTickCount - startTickCount > timeout ) or
         ( FIdleDoneRequested = True ) ) then
    begin
       if ( doneRequested = True ) then
       begin
          // === 2001/12/20 - Exit Idle state ===
          buff_tail := buff_tail + imap_tag + ' BAD - SAKTIMEOUT Error.  Operation quit after: '
               + IntToStr( currentTickCount - startTickCount ) + 'ms.' + #13#10;
          l := copy( buff_tail, pos_tag, length(buff_tail));
          break;
       end;
       cmd := DoneStr + crlf;
       FSendTextToSocket( cmd );
       FIdleDoneRequested := False;
       doneRequested := True;
       startTickCount := currentTickCount;
       timeout := FOperationTimeout;
       for loopcount2 := 0 to 25 do   { === wait approximately 2.5 seconds === }
       begin
          Application.ProcessMessages;
          Sleep(0);
       end;
    end;
  end;

  FIdleDonerequested := False;

  FIMAPError := (pos( imap_tag + ' BAD' , l) > 0);

  FReplyString := s;

  if FIMAPError then
  begin
    result := False;
    FDoErr;
    exit;
  end;

  result := pos( imap_tag + ' OK', UpperCase( FReplyString)) > 0;
  if result then
  begin
    FDoIdle(FSelectedFolder, FMsgsCount, FRecentMsgsCount, True);
  end;

  {$IFDEF SAK_WSOCK_CANUSESOCKSTREAM}
  finally
     if assigned( tempsockstream ) then
        tempsockstream.Free();
     //tempsockstream := nil;
  end;
  {$ENDIF}
end;

function TSakIMAP.LowLevelSearch( const charsetstr: string; const searchcriteria: string;
                                  const UIDsearch: boolean  ) : boolean; // == 2001/12/21 - Search results are msgno
var
  cmd : string;
begin
  result := False;

  if not FConnected then
    raise Exception.create( sak_esNotConnected);

  imap_tag := ImapGetNextTag(imap_tag);

  if ( UIDsearch = True ) then
     cmd := imap_tag + ' UID SEARCH '
  else
     cmd := imap_tag + ' SEARCH ';

  if ( charsetstr <> '' ) then
     cmd := cmd + 'CHARSET ' + charsetstr + ' ';

  cmd := cmd + searchcriteria + crlf;

  Application.ProcessMessages;
  if FCanceled then exit;

  FSendTextToSocket( cmd);
  FReceiveTextFromSocket;
  if FIMAPError then
  begin
    FDoErr;
    exit;
  end;

  result := Pos( imap_tag + ' OK', UpperCase( FReplyString)) > 0;
  if ( result = True ) then
  begin
     if ( UIDsearch = True ) then
        FDoSearch( FSelectedFolder, charsetstr, searchcriteria )
     else
        FDoSearchUID( FSelectedFolder, charsetstr, searchcriteria );
  end;
end;

function TSakIMAP.LowLevelCopy( const MessageSet : string; const FolderName : string;
                                const UIDcopy: boolean ): boolean;  // === 2001/12/21 - UID or msgno
var
  cmd: string;
  tmpFolder : string; // === 2001/12/10
begin
  //result := False;
  imap_tag := ImapGetNextTag(imap_tag);
  tmpFolder := FixFolderName( FolderName );  // === 2001/12/10
  if ( UIDcopy = True ) then
     cmd := imap_tag + ' UID COPY ' + MessageSet + ' ' + tmpFolder +  crlf // === 2001/12/10
  else
     cmd := imap_tag + ' COPY ' + MessageSet + ' ' + tmpFolder +  crlf; // === 2001/12/10

  FSendTextToSocket( cmd);
  FReceiveTextFromSocket;

  if FIMAPError then
  begin
    result := False;
    FDoErr;
    exit;
  end;

  result := pos( imap_tag + ' OK', UpperCase( FReplyString)) > 0;
  if ( result = True ) then
  begin
     if ( UIDcopy = True ) then
        FDoCopyUID( MessageSet, FolderName )
     else
        FDoCopy( MessageSet, FolderName );
  end;

end;

function TSakIMAP.LowLevelStoreFlags( const MsgIndex: string; Flags: TIMAPFlags;
                                      const UIDstore: boolean ): boolean;
// The STORE command alters the Flags associated with the message.
// The \Recent Flag is a read only sstem flag and hence cannot be changed.
var
  cmd, sFlags: string;
  tempindex : word;
begin
  result := False;
  if not FConnected then
    raise Exception.create( sak_esNotConnected);

  imap_tag := ImapGetNextTag(imap_tag);
  If isSeen in Flags then
    sFlags := '\Seen ';
  If isAnswered in Flags then
    sFlags := sFlags + '\Answered ';
  If isFlagged in Flags then
    sFlags := sFlags + '\Flagged ';
  If isDeleted in Flags then
    sFlags := sFlags + '\Deleted ';
  If isDraft in Flags then
    sFlags := sFlags + '\Draft ';

  // The FLAGS argument is used to replace ALL current flags.
  // The SILENT extension is to reduce the response to an OK only.
  if ( UIDstore = True ) then
     cmd := imap_tag + ' UID STORE ' + MsgIndex + ' FLAGS.SILENT (' + sFlags + ')' + crlf
  else
     cmd := imap_tag + ' STORE ' + MsgIndex + ' FLAGS.SILENT (' + sFlags + ')' + crlf;

  Application.ProcessMessages;
  if FCanceled then exit;

  FSendTextToSocket( cmd);
  FReceiveTextFromSocket;
  if FIMAPError then
  begin
    FDoErr;
    exit;
  end;

  result := pos( imap_tag + ' OK', UpperCase( FReplyString)) > 0;

  if result then    // === 2001/03/22 - Listen event added - James.
  begin
     if ( UIDstore = True ) then
        FDoStoreFlagsUID( MsgIndex, Flags )
     else
     begin
        try
           tempindex := StrToInt( MsgIndex );
        except
           tempindex := 0;
        end;
        FDoStoreFlags( tempindex, Flags );
     end;
  end;
end;

// === 2001/12/19 - Search Request - selected folder ===
function TSakIMAP.Search( const charsetstr: string; const searchcriteria: string ) : boolean;
begin
  result := LowLevelSearch( charsetstr, searchcriteria, False );
end;

// === 2001/12/19 - Search by UID Request - selected folder ===
function TSakIMAP.UIDSearch( const charsetstr: string; const searchcriteria: string ) : boolean;
begin
  result := LowLevelSearch( charsetstr, searchcriteria, True );
end;

// === 2001/12/19 - Store Flags by UID - selected folder ===
function TSakIMAP.UIDStoreFlags( const UIDstr: string; Flags: TIMAPFlags ) : boolean;
begin
  result := LowLevelStoreFlags( UIDstr, Flags, True );
end;

// === 2001/12/19 - Fetch by UID - selected folder ===
function TSakIMAP.UIDFetch( const UIDset: string; elements: TIMAPFetchList ) : boolean;
begin
  result := LowLevelFetch( UIDset, elements, False, True );
end;

// === 2001/12/19 - Copy by UID - selected folder ===
function TSakIMAP.UIDCopyMessageToFolder(const UIDSet : string; const FolderName : string ) : boolean;
begin
  result := LowLevelCopy( UIDset, FolderName, True );
end;

// === 2001/12/21 - Authenticate - Plain only currently ===
function TSakIMAP.Authenticate( authtype: TIMAPAuthenticateList; const paramList : TStringList ) : boolean; // === 2001/12/21 - Plain only
const
  CommandContinue : string = '+';
var
  cmd: string;
  authsuccess : boolean;
  receiveresult : boolean;
  authEncodeStream : TBase64EncodingStream;
  authEncodeMemory : TMemoryStream;
  authString : string;
begin
  result := False;
  authEncodeMemory := nil;
  authEncodeStream := nil;
  if not FConnected then
  begin
    raise Exception.create( 'Invalid Op. Not connected.');
  end;
  authsuccess := False;
  if ( imapAuthPlain in authtype ) then
  begin
     exclude( authtype, imapAuthPlain );
     imap_tag := ImapGetNextTag(imap_tag);
     cmd := imap_tag + ' AUTHENTICATE PLAIN' + crlf;
     FSendTextToSocket( cmd);
     receiveresult := FReceiveTextFromSocketUntilToken( CommandContinue );
     if FIMAPError then
     begin
       FDoErr;
       exit;
     end;
     if ( receiveresult = True ) then
     begin
        try
           authString := '';
           authEncodeMemory := TMemoryStream.Create();
           authEncodeStream := TBase64EncodingStream.Create( authEncodeMemory );
           authString := #0 + UserId + #0 + UserPasswd;
           authEncodeStream.Write(authString[1], Length( authString ));
           authEncodeStream.Free();
           authEncodeStream := nil;
           authEncodeMemory.Seek(0, soFromBeginning);
           authString := '';
           SetLength ( authString, authEncodeMemory.Size );
           authEncodeMemory.Read(authString[1], authEncodeMemory.Size);
           authEncodeMemory.Free();
           authEncodeMemory := nil;
           cmd := authString + crlf;
           FSendTextToSocket( cmd);
           FReceiveTextFromSocket;
           if FIMAPError then
           begin
              FDoErr;
              exit;
           end;
           result := pos( imap_tag + ' OK', UpperCase( FReplyString)) > 0;
           if ( result = True ) then
              authsuccess := True;
           FValidated := result;
        finally
           if assigned( authEncodeStream ) then
           begin
              authEncodeStream.Free();
              //authEncodeStream := nil;
           end;
           if assigned( authEncodeMemory ) then
           begin
              authEncodeMemory.Free();
              //authEncodeMemory := nil;
           end;
        end;
     end
     else
     begin
        authsuccess := False;
     end;
  end;
  if ( ( authsuccess = False ) and ( imapAuthNone in authtype ) ) then
  begin
     exclude( authtype, imapAuthNone );
     result := Login();   // === Fallback ===
  end;
end;

// 2002/01/02 - Namespace query to server ( RFC2342 extension )
function TSakIMAP.Namespace() : boolean;
var
  cmd: string;
begin
  //result := False;
  imap_tag := ImapGetNextTag(imap_tag);
  cmd := imap_tag + ' NAMESPACE' + crlf;

  FSendTextToSocket(cmd);
  FReceiveTextFromSocket;

  if FIMAPError then
  begin
    result := False;
    FDoErr;
    exit;
  end;

  result := pos( imap_tag + ' OK', UpperCase( FReplyString)) > 0;
  if ( result = True ) then
  begin
    FDoNamespace();
  end;
end;

// === 2002/01/02 - Explicit List command - replace older calls
function TSakIMAP.List( const referencename : string; const mailboxwildcard : string ) : boolean;  // === 2002/01/02 - List command
var
  cmd: string;
begin
  //result := False;
  imap_tag := ImapGetNextTag(imap_tag);
  cmd := imap_tag + ' LIST ' + referencename + ' ' + mailboxwildcard + crlf;

  FSendTextToSocket(cmd);
  FReceiveTextFromSocket;

  if FIMAPError then
  begin
    result := False;
    FDoErr;
    exit;
  end;

  result := pos( imap_tag + ' OK', UpperCase( FReplyString)) > 0;
  if ( result = True ) then
  begin
    FDoList( referencename, mailboxwildcard );
  end;
end;

// === 2002/01/02 - Set the default list wildcard based on FListFullHierarchy. ===
procedure TSakIMAP.SetFullHierarchy( const setval : boolean ); // === 2002/01/02 - Set List token
const
   fullliststr : string = '"*"';
   hierarchyliststr : string = '"%"';
begin
   if ( setval = True ) then
   begin
      FDefaultListWildcard := fullliststr;  // ==== 2002/01/02 - Default List wildcard - *
      FListFullHierarchy := True;           // ==== 2002/01/02 - List uses * by default
   end
   else
   begin
      FDefaultListWildcard := hierarchyliststr;  // ==== 2002/01/02 - Default List wildcard - %
      FListFullHierarchy := False;           // ==== 2002/01/02 - List uses % by default
   end;
end;

// === 2001/12/21 - Receive text from the socket until the buffer contains token followed by
//        0 or more characters and then the end of line CRLF.  Normal tag return also finishes.
// ===    Output: Result = True if token found in received text, False if not.
function TSakIMAP.FReceiveTextFromSocketUntilToken( const token: string ) : boolean;
const
   ProcSleepModValue : LongInt = 5;  // === 2001/12/10 - Reduce ProcessMessage requests add a sleep.
   MaxTokenSize : LongInt = IMAP_BUFFER_SIZE div 2;
var
  s, buff_tail: string;
  x, lbt, copystart, pos_tag, pos_token: integer;
  buff: array[1..IMAP_BUFFER_SIZE] of char;
  startTickCount : LongInt;          // === 2001/12/10 - Track Start of operation
  currentTickCOunt : LongInt;        // === 2001/12/10 - Track Current ticks
  loopCount, tokenLength : LongInt;  // === 2001/12/10 - Track Number of loop elements.
  tokenFound, tagFound : boolean;
{$IFDEF SAK_WSOCK_CANUSESOCKSTREAM}
  tempsockstream : TWinSocketStream;
{$ENDIF}
begin
  result := False;
  tokenFound := False;
  tagFound := False;
  s := '';
  buff_tail := '';
  loopCount := 0;
  tokenLength := Length( token );

  if ( ( token = '' ) or ( tokenLength > MaxTokenSize )  ) then
  begin
     //result := False;
     raise Exception.create( 'Invalid Op. Token missing or too large.');
     exit;
  end;

  {$IFDEF SAK_WSOCK_CANUSESOCKSTREAM}
  tempsockstream := nil;
  try
     tempsockstream := TWinSocketStream.Create( FSocket.Socket, IMAP_MAXSOCKSTREAMWAIT );
  {$ENDIF}

  startTickCount := GetTickCount();

  while True do
  begin
    loopCount := loopCount + 1;
    if ( (loopCount mod ProcSleepModValue) = 0 ) then
    begin
       Application.ProcessMessages;
       Sleep(0);
    end;

    {$IFDEF SAK_WSOCK_CANUSESOCKSTREAM}
       if ( tempsockstream.WaitForData(IMAP_MAXSOCKSTREAMWAIT) = True ) then
          x := FSocket.Socket.ReceiveBuf( buff, IMAP_BUFFER_SIZE)
       else
       begin
          x := 0; // === No data received.
          if ( FSocket.Socket.Connected = False ) then
             raise Exception.create( 'Receive Failed.  No Longer Connected.' );
       end;
    {$ELSE}
       x := FSocket.Socket.ReceiveBuf( buff, IMAP_BUFFER_SIZE);
    {$ENDIF}

    if ( x <= 0 ) then
    begin
       Application.ProcessMessages;
       Sleep(0);
    end
    else
    begin
       startTickCount := GetTickCount();
       s := s + copy( buff, 1, x);
       buff_tail := buff_tail + copy( buff, 1, x);
       pos_tag := pos( imap_tag, buff_tail);
       pos_token := pos( token, buff_tail);
       if ( pos_tag > 0 ) then
       begin
         tagFound := True;
         if ( pos(crlf, buff_tail) > 0 ) then
           break;
       end;
       if ( tagFound = False ) then
       begin
          if ( pos_token > 0 ) then
          begin
             tokenFound := True;
             lbt := length(buff_tail);
             buff_tail := buff_tail + copy( buff, pos_tag + tokenLength, lbt);
          end
          else
          begin
             lbt := length(buff_tail);
             copystart := lbt - tokenLength + 2;
             if ( copystart > 1 ) then
                buff_tail := copy(buff_tail, copystart, lbt);
          end;
          if ( tokenFound = True ) then
          begin
             if pos(crlf, buff_tail) > 0 then
             begin
                result := True;
                break;
             end;
          end;
       end;
    end; // if x<=0
    currentTickCount := GetTickCount();
    if ( currentTickCount - startTickCount > FOperationTimeout ) then
    begin
       result := False;
       tagFound := True;
       // === 2001/12/10 - Force a BAD return to ensure we exit with an error ===
       buff_tail := buff_tail + imap_tag + ' BAD - SAKTIMEOUT Error.  Operation quit after: '
            + IntToStr( currentTickCount - startTickCount ) + 'ms.' + #13#10;
       break;
    end;
  end;
  if ( tagFound = True ) then
     FIMAPError := (pos( imap_tag + ' BAD' , buff_tail) > 0)
  else
     FIMAPError := False;
  FReplyString := s;
  if assigned( FOnReceiveTextFromSocket) then
  begin
    FOnReceiveTextFromSocket( Self, s);
  end;

  {$IFDEF SAK_WSOCK_CANUSESOCKSTREAM}
  finally
     if assigned( tempsockstream ) then
        tempsockstream.Free();
     //tempsockstream := nil;
  end;
  {$ENDIF}
end;

procedure TSakIMAP.FReceiveTextFromSocket;
const
   ProcSleepModValue : LongInt = 5;  // === 2001/12/10 - Reduce ProcessMessage requests add a sleep.
var
  s, l, buff_tail: string;      // === 2001/12/20 - Use FReceiveMailFromSocket() lbt processing
  x, lbt, pos_tag: integer;     // === 2001/12/20 - Use FReceiveMailFromSocket() lbt processing
  buff: array[1..IMAP_BUFFER_SIZE] of char;
  startTickCount : LongInt;     // === 2001/12/10 - Track Start of operation
  currentTickCOunt : LongInt;   // === 2001/12/10 - Track Current ticks
  loopCount : LongInt;          // === 2001/12/10 - Track Number of loop elements.
{$IFDEF SAK_WSOCK_CANUSESOCKSTREAM}
  tempsockstream : TWinSocketStream;
{$ENDIF}
begin
  // ==== 2001/12/10 - Possible infinite loop conditions in FReceiveTextFromSocket and
  // ====    FReceiveMailFromSocket found and mitigated with a timeout value ( user defined ).
  s := '';
  buff_tail := '';
  loopCount := 0;
  pos_tag := 0;

  {$IFDEF SAK_WSOCK_CANUSESOCKSTREAM}
  tempsockstream := nil;
  try
     tempsockstream := TWinSocketStream.Create( FSocket.Socket, IMAP_MAXSOCKSTREAMWAIT );
  {$ENDIF}

  startTickCount := GetTickCount();

  while True do
  begin
    loopCount := loopCount + 1;
    if ( (loopCount mod ProcSleepModValue) = 0 ) then
    begin
       Application.ProcessMessages;
       Sleep(0);
    end;

    {$IFDEF SAK_WSOCK_CANUSESOCKSTREAM}
       if ( tempsockstream.WaitForData(IMAP_MAXSOCKSTREAMWAIT) = True ) then
          x := FSocket.Socket.ReceiveBuf( buff, IMAP_BUFFER_SIZE)
       else
       begin
          x := 0; // === No data received.
          if ( FSocket.Socket.Connected = False ) then
             raise Exception.create( 'Receive Failed.  No Longer Connected.' );
       end;
    {$ELSE}
       x := FSocket.Socket.ReceiveBuf( buff, IMAP_BUFFER_SIZE);
    {$ENDIF}

    if ( x <= 0 ) then
    begin
       Application.ProcessMessages;
       Sleep(0);
    end
    else
    begin
       startTickCount := GetTickCount();
       s := s + copy( buff, 1, x);

       // === 2001/12/20 - Use same mail message lbt processing.  Same logic.
       if x > 99 then
         buff_tail := copy( buff, x-99, 100)
       else
         buff_tail := buff_tail + copy( buff, 1, x);

       pos_tag := pos( imap_tag, buff_tail);
       // if we found "imap_tag * crlf" then we have finished
       if pos_tag > 0 then
       begin
         l := copy( buff_tail, pos_tag, length(buff_tail));
         if pos(crlf, l) > 0 then
           break;
       end;
       lbt := length(buff_tail);
       if lbt > 99 then
         buff_tail := copy(buff_tail, lbt-40, lbt);
    end; // if x<=0
    currentTickCount := GetTickCount();
    if ( currentTickCount - startTickCount > FOperationTimeout ) then
    begin
       // === 2001/12/10 - Force a BAD return to ensure we exit with an error ===
       buff_tail := buff_tail + imap_tag + ' BAD - SAKTIMEOUT Error.  Operation quit after: '
            + IntToStr( currentTickCount - startTickCount ) + 'ms.' + #13#10;
       l := copy( buff_tail, pos_tag, length(buff_tail));
       break;
    end;
  end;

  FIMAPError := (pos( imap_tag + ' BAD' , l) > 0);

  FReplyString := s;
  if assigned( FOnReceiveTextFromSocket) then
  begin
    FOnReceiveTextFromSocket( Self, s);
  end;

  {$IFDEF SAK_WSOCK_CANUSESOCKSTREAM}
  finally
     if assigned( tempsockstream ) then
        tempsockstream.Free();
     //tempsockstream := nil;
  end;
  {$ENDIF}
end;

procedure TSakIMAP.FReceiveMailFromSocket( RawMail: TStringList);
const
   ProcSleepModValue : LongInt = 5;  // === 2001/12/10 - Reduce ProcessMessage requests add a sleep.
var
  buff_tail, l, tmpFileName: string;
  x, lbt, pos_tag: integer;
  buff: array[1..IMAP_BUFFER_SIZE] of char;
  tmpMsg: TFileStream;
  startTickCount : LongInt;     // === 2001/12/10 - Track Start of operation
  currentTickCOunt : LongInt;   // === 2001/12/10 - Track Current ticks
  loopCount : LongInt;          // === 2001/12/10 - Track Number of loop elements.
{$IFDEF SAK_WSOCK_CANUSESOCKSTREAM}
  tempsockstream : TWinSocketStream;
{$ENDIF}
begin
  // ==== 2001/12/10 - Possible infinite loop conditions in FReceiveTextFromSocket and
  // ====    FReceiveMailFromSocket found and mitigated with a timeout value ( user defined ).

  tmpFileName := sak_getTempFileName;
  tmpMsg := TFileStream.Create( tmpFileName, fmCreate or
                                             fmOpenReadWrite or
                                             fmShareExclusive);

  buff_tail := '';
  loopCount := 0;
  pos_tag := 0;

  {$IFDEF SAK_WSOCK_CANUSESOCKSTREAM}
  tempsockstream := nil;
  try
     tempsockstream := TWinSocketStream.Create( FSocket.Socket, IMAP_MAXSOCKSTREAMWAIT );
  {$ENDIF}

  startTickCount := GetTickCount();

  while True do
  begin
    loopCount := loopCount + 1;
    if ( (loopCount mod ProcSleepModValue) = 0 ) then
    begin
       Application.ProcessMessages;
       Sleep(0);
    end;

    {$IFDEF SAK_WSOCK_CANUSESOCKSTREAM}
       if ( tempsockstream.WaitForData(IMAP_MAXSOCKSTREAMWAIT) = True ) then
          x := FSocket.Socket.ReceiveBuf( buff, IMAP_BUFFER_SIZE)
       else
       begin
          x := 0; // === No data received.
          if ( FSocket.Socket.Connected = False ) then
             raise Exception.create( 'Receive Failed.  No Longer Connected.' );
       end;
    {$ELSE}
       x := FSocket.Socket.ReceiveBuf( buff, IMAP_BUFFER_SIZE);
    {$ENDIF}

    if ( x <= 0 ) then
    begin
       Application.ProcessMessages;
       Sleep(0);
    end
    else
    begin
       startTickCount := GetTickCount();
       try
          tmpMsg.Write( buff, x);
       except
         exit;
       end;

       if x > 99 then
         buff_tail := copy( buff, x-99, 100)
       else
         buff_tail := buff_tail + copy( buff, 1, x);

       pos_tag := pos( imap_tag, buff_tail);
       // if we found "imap_tag * crlf" then we have finished
       if pos_tag > 0 then
       begin
         l := copy( buff_tail, pos_tag, length(buff_tail));
         if pos(crlf, l) > 0 then
           break;
       end;
       lbt := length(buff_tail);
       if lbt > 99 then
         buff_tail := copy(buff_tail, lbt-40, lbt);
    end; // if x<=0

    currentTickCount := GetTickCount();
    if ( currentTickCount - startTickCount > FOperationTimeout ) then
    begin
       // === 2001/12/10 - Force a BAD return to ensure we exit with an error ===
       buff_tail := buff_tail + imap_tag + ' BAD - SAKTIMEOUT Error.  Operation quit after: '
            + IntToStr( currentTickCount - startTickCount ) + 'ms.' + #13#10;
       l := copy( buff_tail, pos_tag, length(buff_tail));
       break;
    end;

  end;

  FIMAPError := (pos( imap_tag + ' BAD' , l) > 0);

  tmpMsg.Position := 0;
  RawMail.LoadFromStream( tmpMsg);
  tmpMsg.Free;
  sysutils.DeleteFile( tmpFileName);

  if RawMail.Count >= 3 then
  begin
    FReplyString := RawMail.Strings[RawMail.Count-1];
    RawMail.Delete(0);
    RawMail.Delete( RawMail.Count-1);
    RawMail.Delete( RawMail.Count-1);
  end;

  if assigned( FOnReceiveTextFromSocket) then
  begin
    FOnReceiveTextFromSocket( Self, FReplyString);
  end;

  {$IFDEF SAK_WSOCK_CANUSESOCKSTREAM}
  finally
     if assigned( tempsockstream ) then
        tempsockstream.Free();
     // tempsockstream := nil;
  end;
  {$ENDIF}
end;

procedure TSakIMAP.FSendTextToSocket( const txt: string);
begin
  Application.ProcessMessages;

  if ( FSocket.Socket.Connected = False ) then
     raise Exception.create( 'Send Failed.  No Longer Connected.' );
  FSocket.Socket.SendBuf( pointer(txt)^, length( txt));
  if assigned( FOnSendTextToSocket) then
  begin
    FOnSendTextToSocket( self, txt);
  end;
end;

procedure TSakIMAP.FDoConnect( Sender: TObject; Socket: TCustomWinSocket);
begin
  FConnected := True;
  if assigned( FOnConnect) then
    FOnConnect( Sender);
end;

procedure TSakIMAP.FDoDisconnect( Sender: TObject; Socket: TCustomWinSocket);
begin
  FConnected := False;
  if assigned( FOnDisconnect) then
     FOnDisconnect( Sender);
end;

procedure TSakIMAP.FDoErr;
var
  ErrorCode: integer;
begin
  ErrorCode := SAK_IMAP_ERROR_CONST;     // === 2001/12/20 - Now use a constant
  FDoError( self, FSocket.Socket, eeGeneral, ErrorCode);
end;

procedure TSakIMAP.FDoError( Sender: TObject; Socket: TCustomWinSocket;
                             ErrorEvent: TErrorEvent; var ErrorCode: Integer);
begin
  FIMAPError := True;
  if ( FCloseOnError = True ) then   // === 2001/12/20 - Flag on/off - James ===
  begin
     FSocket.close;
     FConnected := False;
  end
  else if ( ErrorCode <> SAK_IMAP_ERROR_CONST ) then
  begin
     FSocket.close;
     FConnected := False;
  end;
  if assigned( FOnError) then
  begin
    FOnError( Sender, ErrorCode, FReplyString);
  end;
end;

procedure TSakIMAP.FDoBeforeRetrieve( const MsgIndex: word);
begin
  if assigned( FOnBeforeRetrieve) then
     FOnBeforeRetrieve( self, FSelectedFolder, MsgIndex);
end;

procedure TSakIMAP.FDoAfterRetrieve;
begin
  if assigned( FOnAfterRetrieve) then
     FOnAfterRetrieve( self);
end;

procedure TSakIMAP.FDoDelete( const MsgIndex: word );
begin
  if assigned( FOnDelete) then
     FOnDelete( self, FSelectedFolder, MsgIndex );
end;

procedure TSakIMAP.FDoLookup( Sender: TObject; Socket: TCustomWinSocket);
begin
  if assigned( FOnLookup) then
     FOnLookup( Sender, Socket);
end;

procedure TSakIMAP.FDoConnecting( Sender: TObject; Socket: TCustomWinSocket);
begin
  if assigned( FOnConnecting) then
     FOnConnecting( Sender, Socket);
end;

procedure TSakIMAP.FDoSetFlag( const MsgIndex: word; setFlag : TIMAPFlag );
begin
  if assigned( FOnSetFlag ) then
     FOnSetFlag( self, FSelectedFolder, MsgIndex, setFlag );
end;

procedure TSakIMAP.FDoClearFlag( const MsgIndex: word; setFlag : TIMAPFlag );
begin
  if assigned( FOnClearFlag ) then
     FOnClearFlag( self, FSelectedFolder, MsgIndex, setFlag );
end;

procedure TSakIMAP.FDoStoreFlags( const MsgIndex: word; Flags: TIMAPFlags );
begin
  if assigned( FOnStoreFlags ) then
     FOnStoreFlags( self, FSelectedFolder, MsgIndex, Flags );
end;

procedure TSakIMAP.FDoCheckpoint();
begin
  if assigned( FOnCheckpoint ) then
     FOnCheckpoint( self, FSelectedFolder );
end;

procedure TSakIMAP.FDoSetACL( const folder: string; const identifier: string; rights : TIMAPACLRights );
begin
  if assigned( FOnSetACL ) then
     FOnSetACL( self, folder, identifier, rights );
end;

procedure TSakIMAP.FDoClearACL( const folder: string; const identifier: string; rights : TIMAPACLRights );
begin
  if assigned( FOnSetACL ) then
     FOnClearACL( self, folder, identifier, rights );
end;

procedure TSakIMAP.FDoDeleteACL( const folder : string; const identifier: string );
begin
  if assigned( FOnDeleteACL ) then
     FOnDeleteACL( self, folder, identifier );
end;

procedure TSakIMAP.FDoGetACL( const folder : string );
begin
  if assigned( FOnGetACL ) then
     FOnGetACL( self, folder );
end;

procedure TSakIMAP.FDoListRightsACL( const folder : string; const identifier: string );
begin
  if assigned( FOnListrightsACL ) then
     FOnListrightsACL( self, folder, identifier );
end;

procedure TSakIMAP.FDoMyRightsACL( const folder : string );
begin
  if assigned( FOnMyrightsACL ) then
     FOnMyrightsACL( self, folder );
end;

// ==== 2001/12/18 - Attempt to round out IMAP4rev1 command capabilities ===
procedure TSakIMAP.FDoCapability();
begin
  if assigned( FOnCapability ) then
     FOnCapability( self );
end;

// ==== 2001/12/18 - Attempt to round out IMAP4rev1 command capabilities ===
procedure TSakIMAP.FDoNoop( const folder : string; const numMsgs: word; const recentMsgs: word  );
begin
  if assigned( FOnNoop ) then
     FOnNoop( self, folder, numMsgs, recentMsgs );
end;

// ==== 2001/12/19 - Attempt to round out IMAP4rev1 command capabilities ===
procedure TSakIMAP.FDoStatus( const folder : string; elements: TIMAPStatusList );
begin
  if assigned( FOnStatus ) then
     FOnStatus( self, folder, elements );
end;

// ==== 2001/12/19 - Attempt to round out IMAP4rev1 command capabilities ===
procedure TSakIMAP.FDoFetch( const folder : string; const messageset: string; elements: TIMAPFetchList );
begin
  if assigned( FOnFetch ) then
     FOnFetch( self, folder, messageset, elements );
end;

// ==== 2001/12/19 - Attempt to round out IMAP4rev1 command capabilities ===
procedure TSakIMAP.FDoFetchBody( const folder : string; const messageset: string; elements: TIMAPFetchList;
                            const sectionstring : string; const partialstring : string );
begin
  if assigned( FOnFetchBody ) then
     FOnFetchBody( self, folder, messageset, elements, sectionstring, partialstring );
end;

// ==== 2001/12/21 - Attempt to round out IMAP4rev1 command capabilities ===
procedure TSakIMAP.FDoIdle( const folder : string; const numMsgs: word; const recentMsgs: word; const idleDone: boolean );
begin
  if assigned( FOnIdle ) then
     FOnIdle( self, folder, numMsgs, recentMsgs, idleDone );
end;

// ==== 2001/12/21 - Attempt to round out IMAP4rev1 command capabilities ===
procedure TSakIMAP.FDoStoreFlagsUID( const UIDstr: string; Flags: TIMAPFlags );
begin
  if assigned( FOnStoreFlagsUID ) then
     FOnStoreFlagsUID( self, FSelectedFolder, UIDstr, Flags );
end;

// ==== 2001/12/21 - Attempt to round out IMAP4rev1 command capabilities ===
procedure TSakIMAP.FDoFetchUID( const folder : string; const UIDset: string; elements: TIMAPFetchList );
begin
  if assigned( FOnFetchUID ) then
     FOnFetchUID( self, folder, UIDset, elements );
end;

// ==== 2001/12/21 - Attempt to round out IMAP4rev1 command capabilities ===
procedure TSakIMAP.FDoSearch( const folder : string; const charsetstr: string; const searchcriteria: string );
begin
  if assigned( FOnSearch ) then
     FOnSearch( self, folder, charsetstr, searchcriteria );
end;

// ==== 2001/12/21 - Attempt to round out IMAP4rev1 command capabilities ===
procedure TSakIMAP.FDoSearchUID( const folder : string; const charsetstr: string; const searchcriteria: string );
begin
  if assigned( FOnSearchUID ) then
     FOnSearchUID( self, folder, charsetstr, searchcriteria );
end;

// ==== 2001/12/21 - Attempt to round out IMAP4rev1 command capabilities ===
procedure TSakIMAP.FDoCopy( const MessageSet: string; const FolderName: string );
begin
  if assigned( FOnCopy ) then
     FOnCopy( self, MessageSet, FolderName );
end;

// ==== 2001/12/21 - Attempt to round out IMAP4rev1 command capabilities ===
procedure TSakIMAP.FDoCopyUID( const UIDset: string; const FolderName: string );
begin
  if assigned( FOnCopyUID ) then
     FOnCopyUID( self, UIDset, FolderName );
end;

// ==== 2002/01/02 - RFC2342 extension event ===
procedure TSakIMAP.FDoNamespace();
begin
  if assigned( FOnNamespace) then
    FOnNamespace( self );
end;

// ==== 2002/01/02 - List event. ===
procedure TSakIMAP.FDoList( const referencename : string; const mailboxwildcard : string );
begin
   if assigned( FOnList ) then
      FOnList( self, referencename, mailboxwildcard );
end;

function TSakIMAP.ImapGetNextTag(const s: string): string;
begin
  if s = '' then
  begin
    result := 'SAK___1';
  end else
  begin
    result := 'SAK___' + IntToStr( StrToInt( Copy( s, 7, length(s))) +1 );
  end;
end;


// ==== 2001/12/10 - Force the Connection To Abend and reset state - James.
function TSakIMAP.DoForceAbend : boolean;
begin
   //result := False;
   try
      if assigned( FSocket ) then
         if assigned( FSocket.Socket ) then
            FSocket.Socket.Close();
      FConnected := False;
      FValidated := False;
      FReceivingMsg := False;
      result := True;
   except
      on E: Exception do
      begin
         result := False;
      end;
   end;
end;

end.
